from __future__ import annotations

import logging
import os
import re
from contextlib import suppress
from pathlib import Path
from textwrap import dedent
from typing import ClassVar, overload

import dotenv
from rich.table import Table

from vectice.api import Client
from vectice.api.http_error_handlers import InvalidIdError
from vectice.api.json.iteration import IterationOutput
from vectice.api.json.last_assets import ActivityTargetType
from vectice.api.json.phase import PhaseOutput
from vectice.api.json.project import ProjectOutput
from vectice.api.json.workspace import WorkspaceOutput
from vectice.models.iteration import Iteration
from vectice.models.phase import Phase
from vectice.models.project import Project
from vectice.models.representation.dataset_representation import DatasetRepresentation
from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
from vectice.models.representation.model_representation import ModelRepresentation
from vectice.models.representation.model_version_representation import ModelVersionRepresentation
from vectice.models.workspace import Workspace
from vectice.utils.common_utils import _temp_print, hide_logs
from vectice.utils.configuration import Configuration
from vectice.utils.last_assets import _connection_logging, _get_last_assets, _get_last_user_and_default_workspace
from vectice.utils.logging_utils import CONNECTION_PROJECT_LOGGING, CONNECTION_WORKSPACE_LOGGING, format_description
from vectice.utils.vectice_ids_regex import (
    DATASET_VERSION_VID_REG,
    DATASET_VID_REG,
    ITERATION_VID_REG,
    MODEL_VERSION_VID_REG,
    MODEL_VID_REG,
    PHASE_VID_REG,
    PROJECT_VID_REG,
    WORKSPACE_VID_REG,
)

_logger = logging.getLogger(__name__)
DEFAULT_HOST = "https://app.vectice.com"
CAN_NOT_BE_EMPTY_ERROR_MESSAGE = "%s can not be empty."


class Connection:
    """Connect to the Vectice backend (application).

    The Connection class encapsulates a connection to the Vectice App.
    Thus, it authenticates and connects to Vectice.
    This allows you to start interacting with your Vectice assets.

    A Connection can be initialized in three ways:

    1. Passing the relevant arguments to authenticate and connect to Vectice:

        ```python
        import vectice

        connection = vectice.connect(
            api_token="API_TOKEN_FROM_VECTICE",
            host="https://app.vectice.com",
        )
        ```

    2. Passing the path to a configuration file:

        ```python
        import vectice

        connection = vectice.connect(config="vectice_config.json")
        ```

    3. Letting Vectice find the configuration file in specific locations:

        ```python
        import vectice

        connection = vectice.connect()
        ```

    See [`Connection.connect`][vectice.connection.Connection.connect] for more info.
    """

    USER_FILES_PATH: ClassVar[list[str]] = [
        ".vectice",
        str(Path.home() / ".vectice"),
        ".env",
        str(Path.home() / ".env"),
        "/etc/vectice/api.cfg",
    ]

    def __init__(self, api_token: str, host: str):
        """Initialize a connection.

        Parameters:
            api_token: Your private api token.
            host: The address of the Vectice application.

        Raises:
            RuntimeError: When the API and backend versions are incompatible.
        """
        logging.getLogger("Client").propagate = True
        self._workspace: Workspace | None = None
        self._client = Client(token=api_token, api_endpoint=host)
        compatibility = self._client.check_compatibility()
        if compatibility.status != "OK":
            if compatibility.status == "Error":
                _logger.error(f"compatibility error: {compatibility.message}")
                raise RuntimeError(f"compatibility error: {compatibility.message}")
            else:
                _logger.warning(f"compatibility warning: {compatibility.message}")

    def __repr__(self) -> str:
        return (
            "Connection("
            + f"workspace={self._workspace.name if self._workspace else 'None'}, "
            + f"host={self._client.auth.api_base_url}, "
        )

    @property
    def version_api(self) -> str:
        return self._client.version_api

    @property
    def version_backend(self) -> str:
        return self._client.version_backend

    def workspace(self, workspace: str) -> Workspace:
        """Get a workspace.

        Parameters:
            workspace: The id or the name of the desired workspace.

        Returns:
            The desired workspace.
        """
        result = self._get_workspace_from_str(workspace)
        logging_output = dedent(
            f"""
                        Workspace {result.name!r} successfully retrieved."

                        For quick access to the workspace in the Vectice web app, visit:
                        {self._client.auth._API_BASE_URL}/browse/workspace/{result.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return result

    def _build_workspace_from_output(self, output: WorkspaceOutput) -> Workspace:
        workspace = Workspace(output.id, output.name, output.description)
        workspace.__post_init__(self._client, self)
        return workspace

    def _build_project_from_output(self, output: ProjectOutput) -> Project:
        workspace = self._build_workspace_from_output(output.workspace)
        project = Project(output.id, workspace, output.name, output.description)
        return project

    def _build_phase_from_output(self, output: PhaseOutput) -> Phase:
        if output.project is None:
            raise RuntimeError("Failed to get back project from phase output")
        project = self._build_project_from_output(output.project)
        phase = Phase(output.id, project, output.name, output.index, output.status)
        return phase

    def _build_iteration_from_output(self, output: IterationOutput) -> Iteration:
        if output.phase is None:
            raise RuntimeError("Failed to get back phase from iteration output")
        phase = self._build_phase_from_output(output.phase)
        iteration = Iteration(output.id, output.index, phase, output.status)
        return iteration

    def browse(
        self, asset: str
    ) -> (
        Workspace
        | Project
        | Phase
        | Iteration
        | DatasetRepresentation
        | ModelRepresentation
        | DatasetVersionRepresentation
        | ModelVersionRepresentation
    ):
        """Get an asset.

        Parameters:
            asset: The id of the desired asset.

        Returns:
            The desired asset.
        """
        if re.search(WORKSPACE_VID_REG, asset):
            return self.workspace(asset)
        elif re.search(PROJECT_VID_REG, asset):
            return self.project(asset)
        elif re.search(PHASE_VID_REG, asset):
            return self.phase(asset)
        elif re.search(ITERATION_VID_REG, asset):
            return self.iteration(asset)
        elif re.search(MODEL_VID_REG, asset):
            return self.model(asset)
        elif re.search(MODEL_VERSION_VID_REG, asset):
            return self.model_version(asset)
        elif re.search(DATASET_VID_REG, asset):
            return self.dataset(asset)
        elif re.search(DATASET_VERSION_VID_REG, asset):
            return self.dataset_version(asset)

        raise InvalidIdError("asset", asset)

    def project(self, project: str) -> Project:
        """Get a project.

        Parameters:
            project: The id of the desired project.

        Returns:
            The desired project.
        """
        if not re.search(PROJECT_VID_REG, project):
            raise InvalidIdError("project", project)
        output = self._get_project_from_str(project)
        logging_output = dedent(
            f"""
                Project {output.name!r} successfully retrieved."

                For quick access to the Project in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/project/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return output

    def phase(self, phase: str) -> Phase:
        """Get a phase.

        Parameters:
            phase: The id of the desired phase.

        Returns:
            The desired phase.
        """
        if not re.search(PHASE_VID_REG, phase):
            raise InvalidIdError("phase", phase)
        output = self._client.get_full_phase(phase)

        logging_output = dedent(
            f"""
                        Phase {output.name!r} successfully retrieved."

                        For quick access to the Phase in the Vectice web app, visit:
                        {self._client.auth._API_BASE_URL}/browse/phase/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return self._build_phase_from_output(output)

    def iteration(self, iteration: str) -> Iteration:
        """Get an iteration.

        Parameters:
            iteration: The id of the desired iteration.

        Returns:
            The desired iteration.
        """
        if not re.search(ITERATION_VID_REG, iteration):
            raise InvalidIdError("iteration", iteration)
        output = self._client.get_iteration_by_id(iteration, True)
        logging_output = dedent(
            f"""
                Iteration number {output.index!r} successfully retrieved."

                For quick access to the Iteration in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/iteration/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return self._build_iteration_from_output(output)

    def dataset(self, dataset: str) -> DatasetRepresentation:
        """Get a dataset.

        Parameters:
            dataset: The id of the desired dataset.

        Returns:
            The representation of the desired dataset.
        """
        if not re.search(DATASET_VID_REG, dataset):
            raise InvalidIdError("dataset", dataset)
        output = self._client.get_dataset(dataset)
        logging_output = dedent(
            f"""
                Dataset {output.name!r} successfully retrieved."

                For quick access to the Dataset in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/dataset/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return DatasetRepresentation(output)

    def dataset_version(self, version: str) -> DatasetVersionRepresentation:
        """Get a dataset version.

        Parameters:
            version: The id of the desired dataset version.

        Returns:
            The representation of the desired dataset version.
        """
        if not re.search(DATASET_VERSION_VID_REG, version):
            raise InvalidIdError("dataset_version", version)
        output = self._client.get_dataset_version(version)
        logging_output = dedent(
            f"""
                Dataset version {output.name!r} successfully retrieved."

                For quick access to the Dataset version in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/datasetversion/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return DatasetVersionRepresentation(output)

    def model(self, model: str) -> ModelRepresentation:
        """Get a model.

        Parameters:
            model: The id of the desired model.

        Returns:
            The representation of the desired model.
        """
        if not re.search(MODEL_VID_REG, model):
            raise InvalidIdError("model", model)
        output = self._client.get_model(model)
        logging_output = dedent(
            f"""
                Model {output.name!r} successfully retrieved."

                For quick access to the Model in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/model/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return ModelRepresentation(output)

    def model_version(self, version: str) -> ModelVersionRepresentation:
        """Get a model version.

        Parameters:
            version: The id of the desired model version.

        Returns:
            The representation of the desired model version.
        """
        if not re.search(MODEL_VERSION_VID_REG, version):
            raise InvalidIdError("model_version", version)
        output = self._client.get_model_version(version)
        logging_output = dedent(
            f"""
                Model version {output.name!r} successfully retrieved."

                For quick access to the Model version in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/modelversion/{output.id}"""
        ).lstrip()
        _logger.info(logging_output)
        return ModelVersionRepresentation(output)

    def list_workspaces(self) -> None:
        """Prints a list of workspaces in a tabular format, limited to the first 10 items. A link is provided to view the remaining workspaces.

        Returns:
            None
        """
        workspace_outputs = self._client.list_workspaces()

        rich_table = Table(expand=True, show_edge=False)
        rich_table.add_column("workspace id", justify="left", no_wrap=True, min_width=4, max_width=10)
        rich_table.add_column("name", justify="left", no_wrap=True, max_width=20)
        rich_table.add_column("description", justify="left", no_wrap=True, max_width=50)

        for workspace in workspace_outputs.list:
            rich_table.add_row(str(workspace.id), workspace.name, format_description(workspace.description))

        description = dedent(
            f"""
        There are {workspace_outputs.total} workspaces and a maximum of 10 workspaces are displayed in the table below:"""
        ).lstrip()
        tips = dedent(
            """
        To access your personal workspace, use \033[1mconnection\033[0m.my_workspace
        To access a specific workspace, use \033[1mconnection\033[0m.workspace(Workspace ID)"""
        ).lstrip()
        link = dedent(
            f"""
        For quick access to the list of workspaces in the Vectice web app, visit:
        {self._client.auth._API_BASE_URL}/workspaces"""
        ).lstrip()
        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(tips)
        _temp_print(link)

    @property
    def workspaces(self) -> list[Workspace] | None:
        """List the workspaces to which this connection has access.

        Returns:
            The workspaces to which this connection has access.
        """
        _, default_workspace_id = _get_last_user_and_default_workspace(self._client)
        outputs = self._client.list_workspaces().list
        results: list[Workspace] = []
        for output in outputs:
            workspace = Workspace(id=output.id, name=output.name, description=output.description)
            workspace.__post_init__(self._client, self)
            if output.id == default_workspace_id:
                results.insert(0, workspace)
                continue
            results.append(workspace)
        return results

    @property
    def my_workspace(self) -> Workspace:
        """Retrieve your personal workspace.

        Returns:
            Personal workspace.
        """
        asset = self._client.get_user_and_default_workspace()
        if not asset.get("defaultWorkspace"):
            raise ValueError("Default workspace is not set.")
        return self.workspace(asset["defaultWorkspace"]["vecticeId"])

    @property
    def last_workspace(self) -> Workspace:
        """Retrieve last workspace with activity.

        Returns:
            Last workspace with activity.
        """
        target_types = [activity for activity in ActivityTargetType]
        asset = _get_last_assets(target_types, self._client, _logger)
        if not asset or not asset.get("workspace"):
            raise ValueError("Workspace with activity not found.")
        return self.workspace(asset["workspace"]["vecticeId"])

    @overload
    @staticmethod
    def connect(  # type: ignore[misc]
        api_token: str | None = None,
        host: str | None = None,
        config: str | None = None,
        workspace: str | None = None,
        project: None = None,
    ) -> Connection | Workspace | Project:
        ...

    @overload
    @staticmethod
    def connect(
        api_token: str | None = None,
        host: str | None = None,
        config: str | None = None,
        workspace: str | None = None,
        project: str = "",
    ) -> Project:
        ...

    @staticmethod
    def connect(
        api_token: str | None = None,
        host: str | None = None,
        config: str | None = None,
        workspace: str | None = None,
        project: str | None = None,
    ) -> Connection | Workspace | Project:
        """Method to connect to the Vectice backend (application).

        Authentication credentials are retrieved, in order, from:

        1. keyword arguments
        2. configuration file (`config` parameter)
        3. environment variables
        4. environment files in the following order
            - `.vectice` of the working directory
            - `.vectice` of the user home directory
            - `.env` of the working directory
            - `.env` of the user home directory
            - `/etc/vectice/api.cfg` file

        This method uses the `api_token`, `host`, `workspace`, `project` arguments
        or the JSON config provided. The JSON config file is available from the Vectice
        webapp when creating an API token.

        Parameters:
            api_token: The api token provided by the Vectice webapp.
            host: The backend host to which the client will connect.
                If not found, the default endpoint https://app.vectice.com is used.
            config: A JSON config file containing keys VECTICE_API_TOKEN and
                VECTICE_HOST as well as optionally WORKSPACE and PROJECT.
            workspace: The name or id of an optional workspace to return.
            project: The name or id of an optional project to return.

        Raises:
            ValueError: When a project is specified without a workspace.

        Returns:
            A Connection, Workspace, or Project.
        """
        host = host or Connection._get_host(config)
        api_token = api_token or Connection._get_api_token(host, config)
        workspace = workspace or Connection._get_config_workspace(config)
        project = project or Connection._get_config_project(config)
        connection = Connection(api_token=api_token, host=host)
        user_name, workspace_id = _get_last_user_and_default_workspace(connection._client)
        url = connection._client.auth.api_base_url
        if workspace:
            return connection._log_workspace_or_project(workspace, project, user_name, url)
        _connection_logging(_logger, user_name, url, workspace_id)
        return connection

    @staticmethod
    def _get_host(config: str | None) -> str:
        try:
            return Connection._get_config_item("VECTICE_HOST", config)
        except ValueError:
            # To remove in 23.3.1.0
            try:
                host = Connection._get_config_item("VECTICE_API_ENDPOINT", config)
                _logger.warning(
                    "VECTICE_API_ENDPOINT is deprecated and will be removed in 23.3.1.0, please use VECTICE_HOST instead."
                )
                return host
            except ValueError:
                pass

            _logger.debug(f"No VECTICE_HOST provided. Using default host {DEFAULT_HOST}")
            return DEFAULT_HOST

    @staticmethod
    def _get_api_token(host: str, config: str | None) -> str:
        try:
            return Connection._get_config_item("VECTICE_API_TOKEN", config)
        except ValueError as error:
            raise ValueError(
                f"You must provide the api_token. You can generate them by going to the page {host}/account/api-keys"
            ) from error

    @staticmethod
    def _get_config_workspace(config) -> str | None:
        try:
            return Connection._get_config_item("WORKSPACE", config)
        except ValueError:
            return None

    @staticmethod
    def _get_config_project(config) -> str | None:
        try:
            return str(Connection._get_config_item("PROJECT", config))
        except ValueError:
            return None

    @staticmethod
    def _get_config_item(item_name: str, config_path: str | None) -> str:
        # search in provided config file
        with suppress(KeyError, SyntaxError, TypeError):
            config = Configuration(config_path)  # type: ignore[arg-type]
            if config[item_name]:
                _logger.debug(f"Found {item_name} in {config_path}")
                return config[item_name]
        # search in environment variables
        item = os.environ.get(item_name)
        if item:
            _logger.debug(f"Found {item_name} in environment variables.")
            return item
        # search in user configuration files
        for path in Connection.USER_FILES_PATH:
            with hide_logs("dotenv"):
                item = dotenv.get_key(path, item_name)
            if item:
                _logger.debug(f"Found {item_name} in {path}")
                return item
        raise ValueError(f"Could not find {item_name} in user configuration")

    def _log_workspace_or_project(
        self, workspace: str, project: str | None, user_name: str, host: str
    ) -> Workspace | Project:
        workspace_output: Workspace = self._get_workspace_from_str(workspace)
        if project:
            project_obj = self._log_project(workspace_output, project, user_name, host)
            self._workspace = project_obj.workspace
            _logger.debug(
                f"Successfully authenticated. You'll be working on Project: {project_obj.name!r}, part of Workspace: {project_obj.workspace.name!r}"
            )
            return project_obj

        workspace_obj = self._log_workspace(workspace_output, user_name, host)
        self._workspace = workspace_obj
        _logger.debug(f"Successfully authenticated. Your current Workspace: {workspace_output.name!r}")
        return workspace_obj

    def _log_workspace(self, workspace_output: Workspace, user_name: str, host: str) -> Workspace:
        _logger.info(
            CONNECTION_WORKSPACE_LOGGING.format(
                user=user_name, workspace_name=workspace_output.name, url=host, workspace_id=workspace_output.id
            )
        )
        return workspace_output

    def _log_project(self, workspace_output: Workspace, project: str, user_name: str, host: str) -> Project:
        project_output = self._get_project_from_str(project, workspace_output.id)

        _logger.info(
            CONNECTION_PROJECT_LOGGING.format(
                user=user_name,
                project_name=project_output.name,
                url=host,
                workspace_id=workspace_output.id,
                project_id=project_output.id,
            )
        )
        return project_output

    def _get_workspace_from_str(self, workspace: str) -> Workspace:
        return self._build_workspace_from_output(self._client.get_workspace(workspace))

    def _get_project_from_str(self, project: str, workspace: str | None = None) -> Project:
        return self._build_project_from_output(self._client.get_project(project, workspace))



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from vectice.api.json.phase import PhaseOutput

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationStepArtifact


class DocumentationPageStatus(Enum):
    """Enumeration of the different statuses for documentation pages."""

    NotStarted = "NotStarted"
    InProgress = "Draft"
    Completed = "Completed"


class StepType(Enum):
    """Enumeration of the different types of steps."""

    Step = "UNKNOWN"
    StepDataset = "DATASET"
    StepModel = "MODEL"
    StepNumber = "NUMBER"
    StepString = "STRING"
    StepPlot = "PLOT"
    StepImage = "IMAGE"


class StepInput(dict):
    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def index(self) -> int:
        return int(self["index"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def completed(self) -> bool:
        return bool(self["completed"])

    @property
    def description(self) -> str | None:
        return str(self["description"])

    @property
    def artifacts(self) -> list[IterationStepArtifact]:
        # TODO: refactor to break cyclic import
        from vectice.api.json.iteration import IterationStepArtifact

        return [IterationStepArtifact(artifact) for artifact in self["artifacts"]]

    @property
    def step_type(self) -> StepType:
        return StepType(self["stepType"])


class StepOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "phase" in self:
            self._phase: PhaseOutput = PhaseOutput(**self["phase"])
        else:
            self._phase = None
        if "slug" in self:
            self["slug"] = f"step_{self['slug']}"

    @property
    def id(self) -> int:
        return int(self["id"])

    @property
    def index(self) -> int:
        return int(self["index"])

    @property
    def parent(self) -> PhaseOutput:
        return self._phase

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def completed(self) -> bool:
        return bool(self["completed"])

    @property
    def description(self) -> str | None:
        return str(self["description"])

    @property
    def slug(self) -> str:
        return str(self["slug"])

    @property
    def artifacts_count(self) -> int:
        return int(self["artifactsCount"])

    @property
    def artifacts(self) -> list[IterationStepArtifact]:
        # TODO: refactor to break cyclic import
        from vectice.api.json.iteration import IterationStepArtifact

        return [IterationStepArtifact(artifact) for artifact in self["artifacts"]]

    @property
    def step_type(self) -> StepType:
        return StepType(self["stepType"])



from __future__ import annotations

import logging
import pickle  # nosec
from typing import Any

from vectice.models import Metric, Property
from vectice.models.dataset import TDerivedFrom, _get_derived_from

_logger = logging.getLogger(__name__)


class Model:
    """Represent a wrapped model.

    Iteration steps that represent models require a wrapped model.
    Wrapped models present Vectice with a learned predictor and
    appropriate metadata about that predictor in order to summarize
    the model in views of the iteration.
    """

    def __init__(
        self,
        library: str,
        technique: str,
        metrics: dict[str, int] | list[Metric] | Metric | None = None,
        properties: dict[str, str] | dict[str, int] | list[Property] | Property | None = None,
        name: str | None = None,
        attachments: str | list[str] | None = None,
        predictor: Any = None,
        derived_from: list[TDerivedFrom] | None = None,
    ):
        """Wrap a model (predictor).

        A Vectice Model is a wrapped predictor suitable for assignment
        to a Vectice Step.

        Parameters:
            library: The library used to generate the model.
            technique: The modeling technique used.
            metrics: A dict for example `{"MSE": 1}`.
            properties: A dict, for example `{"folds": 32}`.
            name: The model name. If None, will be auto-generated based on the library and technique.
            attachments: Path of a file that will be attached to the step along with the predictor.
            predictor: The predictor.
            derived_from: List of dataset (or version ids) to link as lineage.
        """
        self._library = library
        self._technique = technique
        self._name = name if name else self._generate_name()
        self._metrics = self._format_metrics(metrics) if metrics else None
        self._properties = self._format_properties(properties) if properties else None
        self._attachments = self._format_attachments(attachments) if attachments else None
        self._predictor = pickle.dumps(predictor)  # nosec
        self._derived_from = _get_derived_from(derived_from)

    def __repr__(self):
        return (
            f"Model(name='{self.name}', library='{self.library}', technique='{self.technique}', "
            f"metrics={self.metrics}, properties={self.properties}, attachments={self.attachments})"
        )

    @property
    def name(self) -> str:
        """The model's name.

        Returns:
            The model's name.
        """
        return self._name

    @name.setter
    def name(self, name):
        """Set the model's name.

        Parameters:
            name: The name of the model.
        """
        self._name = name

    @property
    def predictor(self) -> Any:
        """The model's predictor.

        Returns:
            The model's predictor.
        """
        return pickle.loads(self._predictor)  # noqa: S301

    @predictor.setter
    def predictor(self, predictor: Any):
        """Set the model's predictor.

        Parameters:
            predictor: The predictor of the model.
        """
        self._predictor = pickle.dumps(predictor)

    @classmethod
    def read_predictor_file(cls, path: str) -> Any:
        with open(path, "rb") as fh:
            return pickle.load(fh)  # noqa: S301

    @property
    def library(self) -> str:
        """The name of the library used to generate the model.

        Returns:
            The name of the library used to generate the model.
        """
        return self._library

    @library.setter
    def library(self, library):
        """Set the name of the library used to create the model.

        Parameters:
            library: The name of the library used to create the model.
        """
        self._library = library

    @property
    def technique(self) -> str:
        """The name of the modeling technique used to learn the model.

        Returns:
            The name of the modeling technique used to learn the model.
        """
        return self._technique

    @technique.setter
    def technique(self, technique):
        """Set the name of the modeling technique used to learn the model.

        Parameters:
            technique: The modeling technique used.
        """
        self._technique = technique

    @property
    def metrics(self) -> list[Metric] | None:
        """The model's metrics.

        Returns:
            The model's metrics.
        """
        return self._metrics

    @metrics.setter
    def metrics(self, metrics: dict[str, int] | list[Metric] | Metric | None):
        """Set the model's metrics.

        Parameters:
            metrics: The metrics of the model.
        """
        self._format_metrics(metrics)

    @property
    def properties(self) -> list[Property] | None:
        """The model's properties.

        Returns:
            The model's properties.
        """
        return self._properties

    @properties.setter
    def properties(self, properties: dict[str, str] | list[Property] | Property | None):
        """Set the model's properties.

        Parameters:
            properties: The properties of the model.
        """
        self._format_properties(properties)

    @property
    def attachments(self) -> list[str] | None:
        """The attachments associated with the model.

        Returns:
            The attachments associated with the model.
        """
        return self._attachments

    @attachments.setter
    def attachments(self, attachments: list[str] | str):
        """Attach a file or files to the model.

        Parameters:
            attachments: The filename or filenames of the file or set of files to attach to the model.
        """
        self._attachments = self._format_attachments(attachments)

    @property
    def derived_from(self) -> list[str]:
        """The datasets from which this model is derived.

        Returns:
            The datasets from which this model is derived.
        """
        return self._derived_from

    def _generate_name(self) -> str:
        return f"{self.library} {self.technique} model"

    def _format_metrics(self, metrics: dict[str, int] | list[Metric] | Metric | None) -> list[Metric]:
        if metrics is None:
            return []
        if isinstance(metrics, Metric):
            return [metrics]
        if isinstance(metrics, list):
            metrics = self._remove_incorrect_metrics(metrics)
            key_list = [metric.key for metric in metrics]
            self._check_key_duplicates(key_list)
            return metrics
        if isinstance(metrics, dict):
            return [Metric(key, value) for (key, value) in metrics.items()]
        else:
            raise ValueError("Please check metric type.")

    @staticmethod
    def _check_key_duplicates(key_list: list[str]):
        if len(key_list) != len(set(key_list)):
            raise ValueError("Duplicate keys are not allowed.")

    @staticmethod
    def _remove_incorrect_metrics(metrics: list[Metric]) -> list[Metric]:
        for metric in metrics:
            if not isinstance(metric, Metric):
                logging.warning(f"Incorrect metric '{metric}'. Please check metric type.")
                metrics.remove(metric)
        return metrics

    @staticmethod
    def _format_attachments(attachments: str | list[str]) -> list[str]:
        return [attachment for attachment in set(attachments)] if isinstance(attachments, list) else [attachments]

    def _format_properties(
        self, properties: dict[str, str] | dict[str, int] | list[Property] | Property | None
    ) -> list[Property]:
        if properties is None:
            return []
        if isinstance(properties, Property):
            return [properties]
        if isinstance(properties, list):
            properties = self._remove_incorrect_properties(properties)
            key_list = [prop.key for prop in properties]
            self._check_key_duplicates(key_list)
            return properties
        if isinstance(properties, dict):
            return [Property(key, str(value)) for (key, value) in properties.items()]
        else:
            raise ValueError("Please check property type.")

    @staticmethod
    def _remove_incorrect_properties(properties: list[Property]) -> list[Property]:
        for prop in properties:
            if not isinstance(prop, Property):
                logging.warning(f"Incorrect property '{prop}'. Please check property type.")
                properties.remove(prop)
        return properties



from __future__ import annotations

from enum import Enum


class ActivityTargetType(Enum):
    Code = "Code"
    CodeVersion = "CodeVersion"
    DataResource = "DataResource"
    DataSet = "DataSet"
    DataSetVersion = "DataSetVersion"
    Datasheet = "Datasheet"
    Iteration = "Iteration"
    Review = "Review"
    ReviewComment = "ReviewComment"
    IterationStep = "IterationStep"
    Model = "Model"
    ModelCard = "ModelCard"
    ModelVersion = "ModelVersion"
    Phase = "Phase"
    Project = "Project"
    StepDefinition = "StepDefinition"
    Workspace = "Workspace"
    WorkspaceUser = "WorkspaceUser"


class UserActivity(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def id(self) -> int:
        return int(self["targetId"])

    @property
    def name(self) -> str:
        return str(self["targetName"])

    @property
    def target_type(self) -> ActivityTargetType:
        return ActivityTargetType(self["targetType"])



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

if TYPE_CHECKING:
    from google.cloud.storage import Blob, Client

GS_URI_REG = r"(gs:\/\/)([^\/]+)\/(.+)"

_logger = logging.getLogger(__name__)


class GCSResource(Resource):
    """GCS resource reference wrapper.

    This resource wraps GCS uris references such as file folders that you have stored in Google Cloud
    Storage with optional metadata and versioning. You assign it to a step.

    ```python
    from vectice import GCSResource

    gcs_resource = GCSResource(
        uris="gs://<bucket_name>/<file_path_inside_bucket>",
    )
    ```
    """

    _origin = DatasetSourceOrigin.GCS.value

    def __init__(
        self,
        uris: str | list[str],
        dataframes: DataFrameType | list[DataFrameType] | None = None,
        gcs_client: Client | None = None,
    ):
        """Initialize a GCS resource.

        Parameters:
            uris: The uris of the referenced resources. Should follow the pattern 'gs://<bucket_name>/<file_path_inside_bucket>'
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas)
            gcs_client (Optional): The `google.cloud.storage.Client` to optionally retrieve file size, creation date and updated date (used for auto-versioning) up to 5000 files.
        """
        super().__init__(paths=uris, dataframes=dataframes)
        self.gcs_client = gcs_client

        for uri in self._paths:
            if not re.search(GS_URI_REG, uri):
                raise ValueError(
                    f"Uri '{uri}' is not following the right pattern 'gs://<bucket_name>/<file_path_inside_bucket>'"
                )

    def _fetch_data(self) -> dict[str, bytes | None]:
        datas = {}
        for uri in self._paths:
            bucket_name, path = self._get_bucket_and_path_from_uri(uri)
            blobs = self._get_blobs(bucket_name, path)
            if blobs is not None:
                for blob in blobs:
                    datas[f"{bucket_name}/{path}"] = blob
        return datas

    def _build_metadata(self) -> FilesMetadata:
        files = []
        size: int | None = None
        df_index = 0
        for uri in self._paths:
            bucket_name, path = self._get_bucket_and_path_from_uri(uri)
            blobs = self._get_blobs(bucket_name, path)
            if blobs is not None:
                sorted_blobs = sorted(blobs, key=lambda bl: str(bl.name).lower())
                for blob in sorted_blobs:
                    dataframe = (
                        self._dataframes[df_index]
                        if self._dataframes is not None and len(self._dataframes) > df_index
                        else None
                    )
                    blob_file = self._build_file_from_blob(blob, f"gs://{bucket_name}", dataframe)
                    files.append(blob_file)
                    if size is None and blob_file.size is not None:
                        size = 0
                    if size is not None:
                        size += blob_file.size or 0
                    df_index += 1
            else:
                dataframe = (
                    self._dataframes[df_index]
                    if self._dataframes is not None and len(self._dataframes) > df_index
                    else None
                )
                files.append(File(name=path, uri=uri, dataframe=dataframe))
                df_index += 1
        metadata = FilesMetadata(
            size=size,
            origin=self._origin,
            files=files,
        )
        return metadata

    def _get_blobs(self, bucket_name: str, path: str) -> list[Blob] | None:
        if self.gcs_client is None:
            return None

        blobs_list = list(
            filter(
                lambda bl: bl.name.endswith("/") is False,
                self.gcs_client.list_blobs(
                    bucket_or_name=bucket_name, prefix=path, max_results=(self._files_limit + 2)
                ),  # +2 for base directory and to check if more files than limit
            )
        )

        if len(blobs_list) > self._files_limit:
            _logger.warning(f"Only first {self._files_limit} files metadata were used.")
            blobs_list.pop()

        return blobs_list

    def _build_file_from_blob(self, blob: Blob, uri: str, dataframe: DataFrameType | None = None) -> File:
        return File(
            name=blob.name,
            size=blob.size,
            fingerprint=blob.md5_hash,
            created_date=blob.time_created.isoformat(),
            updated_date=blob.updated.isoformat(),
            uri=f"{uri}/{blob.name}",
            dataframe=dataframe,
            content_type=blob.content_type,
        )

    def _get_bucket_and_path_from_uri(self, uri: str) -> tuple[str, str]:
        match = re.search(GS_URI_REG, uri)
        if match is not None:
            _, bucket_name, path = match.groups()
            return bucket_name, path

        raise ValueError(
            f"Uri '{uri}' is not following the right pattern 'gs://<bucket_name>/<file_path_inside_bucket>'"
        )


class NoSuchGCSResourceError(Exception):
    def __init__(self, bucket: str, resource: str):
        self.message = f"{resource} does not exist in the GCS bucket {bucket}."
        super().__init__(self.message)



from __future__ import annotations

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json import OrgConfigOutput

_RETURNS = """
    organization {
        configuration {
            dfStatisticsRowThreshold
        }
    }
    __typename
"""


class GqlOrganizationApi(GqlApi):
    def get_organization_config(self) -> OrgConfigOutput:
        query = GqlApi.build_query(
            gql_query="getOrgConfig",
            returns=_RETURNS,
            query=True,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built)
            print(response)
            output: OrgConfigOutput = Parser().parse_item(response["getOrgConfig"])
            return output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "code", "getOrgConfig")



from __future__ import annotations

from vectice.models.resource.metadata.base import (
    DatasetSourceType,
    DatasetSourceUsage,
    Metadata,
)
from vectice.models.resource.metadata.column_metadata import DBColumn
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.source import Source


class DBMetadata(Metadata):
    """Class that describes metadata of dataset that comes from a database."""

    def __init__(
        self,
        dbs: list[MetadataDB],
        size: int | None,
        usage: DatasetSourceUsage | None = None,
        origin: str | None = None,
    ):
        """Initialize a DBMetadata instance.

        Parameters:
            dbs: The list of databases.
            size: The size of the metadata.
            usage: The usage of the metadata.
            origin: The origin of the metadata.
        """
        super().__init__(size=size, type=DatasetSourceType.DB, usage=usage, origin=origin)
        self.dbs = dbs

    def asdict(self) -> dict:
        for db in self.dbs:
            if self._settings is not None:
                db.set_settings(self._settings)
        return {
            **super().asdict(),
            "dbs": [db.asdict() for db in self.dbs],
        }


class MetadataDB(Source):
    def __init__(
        self,
        name: str,
        columns: list[DBColumn],
        rows_number: int | None = None,
        size: int | None = None,
        updated_date: str | None = None,
        created_date: str | None = None,
        dataframe: DataFrameType | None = None,
    ):
        """Initialize a MetadataDB instance.

        Parameters:
            name: The name of the table.
            columns: The columns that compose the table.
            rows_number: The number of row of the table.
            size: The size of the table.
            updated_date: The date of last update of the table.
            created_date: The creation date of the table.
            dataframe (Optional): A dataframe allowing vectice to optionally compute more metadata about this resource such as columns stats, size, rows number and column numbers. (Support Pandas)
        """
        super().__init__(
            name=name,
            size=size,
            columns=list(columns),
            updated_date=updated_date,
            created_date=created_date,
            dataframe=dataframe,
        )
        self.rows_number = rows_number

    def asdict(self) -> dict:
        return {
            "rowsNumber": self.rows_number,
            **super().asdict(),
        }



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    pass


@dataclass
class WorkspaceInput:
    name: str | None
    description: str | None


class WorkspaceOutput(dict):
    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None



from __future__ import annotations

import logging

from vectice.api.json.iteration import IterationStepArtifactInput
from vectice.api.json.step import StepType
from vectice.models.step import Step
from vectice.utils.common_utils import _check_read_only

_logger = logging.getLogger(__name__)


class StepNumber(Step):
    """Model a Vectice step's number.

    A StepNumber stores numeric values.
    """

    def __init__(self, step: Step, number: int | float | None = None):
        """Initialize a number step.

        Parameters:
            step: The step.
            number: The step's number.
        """
        super().__init__(
            id=step.id,
            iteration=step._iteration,
            name=step.name,
            index=step.index,
            slug=step.slug,
            description=step._description,
            step_type=StepType.StepNumber,
            artifacts=step.artifacts,
        )
        self._number = number

    def __repr__(self):
        return (
            f"StepNumber(name={self.name!r}, id={self.id!r}, description={self._description!r}, number={self.number!r})"
        )

    @property
    def number(self) -> int | float | None:
        """The step's number.

        Returns:
            The step's number.
        """
        return self._number

    @number.setter
    def number(self, value: int | float) -> None:
        """Set the step's number.

        Typical usage example:

        ```pycon
        >>> my_iteration = my_phase.create_iteration()
        ... my_iteration.step_scoring = 21
        ```

        Parameters:
            value: The number.
        """
        _check_read_only(self.iteration)
        step_artifacts = [IterationStepArtifactInput(type="Comment", text=value)]
        self._client.update_iteration_step_artifact(
            self.id,
            StepType.StepNumber,
            step_artifacts,
        )
        self._number = value



from __future__ import annotations

import logging
import re
from typing import TYPE_CHECKING

from vectice.models.resource.base import Resource
from vectice.models.resource.metadata.base import DatasetSourceOrigin
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.files_metadata import File, FilesMetadata

if TYPE_CHECKING:
    from mypy_boto3_s3 import Client
    from mypy_boto3_s3.type_defs import ListObjectsV2OutputTypeDef, ObjectTypeDef


S3_URI_REG = r"(s3:\/\/)([^\/]+)\/(.+)"


_logger = logging.getLogger(__name__)


class S3Resource(Resource):
    """AWS S3resource reference wrapper.

    This resource wraps AWS S3 uris references such as file folders that you have stored in AWS S3
    with optional metadata and versioning. You assign it to a step.

    ```python
    from vectice import S3Resource

    s3_resource = S3Resource(
        uris="s3://<bucket_name>/<file_path_inside_bucket>",
    )
    ```
    """

    _origin = DatasetSourceOrigin.S3.value

    def __init__(
        self,
        uris: str | list[str],
        dataframes: DataFrameType | list[DataFrameType] | None = None,
        s3_client: Client | None = None,
    ):
        """Initialize an S3 resource.

        Parameters:
            uris: The uris of the resources to get. Should follow the pattern 's3://<bucket_name>/<file_path_inside_bucket>'
            dataframes (Optional): The dataframes allowing vectice to optionally compute more metadata about this resource such as columns stats. (Support Pandas)
            s3_client (Optional): The Amazon s3 client to optionally retrieve file size, creation date and updated date (used for auto-versioning) up to 5000 files.
        """
        super().__init__(paths=uris, dataframes=dataframes)
        self.s3_client = s3_client

        for uri in self._paths:
            if not re.search(S3_URI_REG, uri):
                raise ValueError(
                    f"Uri '{uri}' is not following the right pattern 's3://<bucket_name>/<file_path_inside_bucket>'"
                )

    def _fetch_data(self) -> dict[str, ObjectTypeDef | None]:
        datas: dict[str, ObjectTypeDef | None] = {}
        if self.s3_client:
            for bucket_name, s3_objects_list in self._get_s3_objects_list(self.s3_client):
                for s3_objects in s3_objects_list:
                    for s3_object in s3_objects["Contents"]:
                        object_path = s3_object["Key"]
                        datas[f"{bucket_name}/{object_path}"] = s3_object
        else:
            for path in self._paths:
                datas[path] = None

        return datas

    def _build_metadata(self) -> FilesMetadata:
        size = None
        files = []
        df_index = 0
        if self.s3_client:
            for bucket_name, s3_objects_list in self._get_s3_objects_list(self.s3_client):
                for s3_objects in s3_objects_list:
                    if s3_objects["KeyCount"] == 0:
                        raise NoSuchS3ResourceError(bucket_name, s3_objects["Prefix"])
                    s3_object = s3_objects["Contents"]
                    new_files, total_size, new_df_index = self._build_files_list_with_size(
                        index=df_index, bucket_name=bucket_name, s3_object=s3_object
                    )
                    if size is None:
                        size = 0
                    size += total_size
                    files.extend(new_files)
                    df_index += new_df_index
        else:
            for index, uri in enumerate(self._paths):
                dataframe = (
                    self._dataframes[index] if self._dataframes is not None and len(self._dataframes) > index else None
                )
                _, path = self._get_bucket_and_path_from_uri(uri)
                file = File(name=path, uri=uri, dataframe=dataframe)
                files.append(file)

        metadata = FilesMetadata(files=files, origin=self._origin, size=size)
        return metadata

    def _build_files_list_with_size(
        self, index: int, bucket_name: str, s3_object: list[ObjectTypeDef]
    ) -> tuple[list[File], int, int]:
        files: list[File] = []
        total_size = 0
        df_index = 0
        filtered_s3_object = list(filter(lambda obj: self._is_s3_object_a_folder(obj) is False, s3_object))
        sorted_s3_object = sorted(filtered_s3_object, key=lambda obj: obj["Key"].lower())
        for object in sorted_s3_object:
            new_index = df_index + index
            name = object["Key"]
            size = object["Size"]
            uri = f"s3://{bucket_name}/{name}"
            dataframe = (
                self._dataframes[new_index]
                if self._dataframes is not None and len(self._dataframes) > new_index
                else None
            )

            file = File(
                name=name,
                size=size,
                fingerprint=self._get_formatted_etag_from_object(object),
                updated_date=object["LastModified"].isoformat(),
                created_date=None,
                uri=uri,
                dataframe=dataframe,
            )
            total_size += size
            files.append(file)
            df_index += 1
        return files, total_size, df_index

    def _get_s3_objects_list(self, s3_client: Client) -> list[tuple[str, list[ListObjectsV2OutputTypeDef]]]:
        return list(map(lambda uri: self._get_s3_objects(s3_client, uri), self._paths))

    def _get_s3_objects(self, s3_client: Client, uri: str) -> tuple[str, list[ListObjectsV2OutputTypeDef]]:
        bucket_name, path = self._get_bucket_and_path_from_uri(uri)
        results: list[ListObjectsV2OutputTypeDef] = []
        next_token: str | None = ""
        base_kwargs = {
            "Bucket": bucket_name,
            "Prefix": path,
        }
        count = 0
        while next_token is not None:
            kwargs = base_kwargs.copy()
            if next_token != "":
                kwargs.update({"ContinuationToken": next_token})

            result = s3_client.list_objects_v2(**kwargs)  # type: ignore[arg-type]
            count += result["KeyCount"]
            if count >= self._files_limit:
                _logger.warning(f"Only first {self._files_limit} files metadata were used.")
                next_token = None
            else:
                next_token = result.get("NextContinuationToken")

            results.append(result)

        return bucket_name, results

    def _get_bucket_and_path_from_uri(self, uri: str) -> tuple[str, str]:
        match = re.search(S3_URI_REG, uri)
        if match is not None:
            _, bucket_name, path = match.groups()
            return bucket_name, path

        raise ValueError(
            f"Uri '{uri}' is not following the right pattern 's3://<bucket_name>/<file_path_inside_bucket>'"
        )

    @staticmethod
    def _get_formatted_etag_from_object(object: ObjectTypeDef) -> str:
        return str(object["ETag"].replace("'", "").replace('"', ""))

    @staticmethod
    def _is_s3_object_a_folder(object: ObjectTypeDef) -> bool:
        return bool(object["Key"].endswith("/"))


class NoSuchS3ResourceError(Exception):
    def __init__(self, bucket: str, resource: str):
        self.message = f"{resource} does not exist in the AWS s3 bucket {bucket}."
        super().__init__(self.message)



from __future__ import annotations

from typing import TYPE_CHECKING

from vectice.api.json.step import StepType
from vectice.models.step import Step

if TYPE_CHECKING:
    from vectice.api.json.iteration import IterationStepArtifact
    from vectice.models.model import Model


class StepModel(Step):
    """Model a model step.

    A model step stores a model version id.
    See also [`Dataset`][vectice.models.model.Model].
    """

    def __init__(self, step: Step, model_version: IterationStepArtifact, model: Model | None = None):
        """Initialize a dataset step.

        Parameters:
            step: The step.
            model_version: The step's model version.
        """
        super().__init__(
            step.id,
            step._iteration,
            step.name,
            step.index,
            step.slug,
            step._description,
            step_type=StepType.StepModel,
            artifacts=step.artifacts,
        )
        self._model_version = model_version
        self._model = model

    def __repr__(self):
        return f"StepModel(name={self.name!r}, id={self.id!r}, description={self._description!r}, model_version_id={self.model_version!r})"

    @property
    def model_version(self) -> IterationStepArtifact:
        """The step's model.

        Returns:
            The step's model.
        """
        return self._model_version



from __future__ import annotations

import io
import logging
import os
from typing import TYPE_CHECKING, BinaryIO, List

from vectice.api.json.dataset_version import DatasetVersionOutput
from vectice.api.json.model_version import ModelVersionOutput

if TYPE_CHECKING:
    from vectice.api import Client

_logger = logging.getLogger(__name__)

FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE = "The file path '%s' is not valid. The file does not exist."


class AttachmentContainer:
    def __init__(self, version: ModelVersionOutput | DatasetVersionOutput, client: Client):
        self._client = client
        self._version = version
        self._id = version.id
        self._name = version.name

    @property
    def version(self):
        return self._version

    @property
    def name(self):
        return self._name

    @property
    def id(self):
        return self._id

    def add_attachments(self, file_paths: str | list[str]):
        """Add an attachment or set of attachments to the entity.

        Parameters:
            file_paths: The paths of the attachment(s).

        Returns:
            A list of dictionaries describing the attachments.
        """
        file_paths = [file_paths] if isinstance(file_paths, str) else file_paths
        attachments = self._add_files_to_attachments(file_paths)
        try:
            return self._client.create_version_attachments(attachments, self._version)
        finally:
            self._close_attached_files(attachments)

    def _add_files_to_attachments(self, file_paths: list[str]) -> list[tuple[str, tuple[str, BinaryIO]]]:
        attachments: List[tuple[str, tuple[str, BinaryIO]]] = []
        for file_path in file_paths:
            if not os.path.exists(file_path):
                raise ValueError(FILE_PATH_DOES_NOT_EXIST_ERROR_MESSAGE % file_path)
            curr_file = ("file", (file_path, open(file_path, "rb")))
            attachments.append(curr_file)
        return attachments

    def _close_attached_files(self, attachments: list[tuple[str, tuple[str, BinaryIO]]]) -> None:
        for attachment in attachments:
            attachment[1][1].close()

    def add_serialized_model(self, model_type: str, model_content: bytes):
        if isinstance(self._version, ModelVersionOutput):
            self._client.create_model_predictor(model_type, io.BytesIO(model_content), self._version)



from __future__ import annotations

from textwrap import dedent
from typing import TYPE_CHECKING, Any

from vectice.utils.automatic_link_utils import existing_dataset_logger, existing_model_logger

if TYPE_CHECKING:
    from logging import Logger

    from vectice.api import Client
    from vectice.api.json.last_assets import ActivityTargetType
    from vectice.models.step import Step


def _get_last_assets(target_types: list[ActivityTargetType], client: Client, _logger: Logger):
    target_types_array = [target.value for target in target_types]
    try:
        last_asset = client.get_last_assets(target_types_array, {"index": 1, "size": 1}).list[0]
        return last_asset
    except IndexError as e:
        _logger.debug(f"There were no assets with activity found. Sanity check {e}")
        return


def _get_last_user_and_default_workspace(client: Client) -> tuple[str, str | None]:
    asset = client.get_user_and_default_workspace()
    workspace_id = str(asset["defaultWorkspace"]["vecticeId"]) if asset["defaultWorkspace"] else None
    return asset["user"]["name"], workspace_id


def _connection_logging(_logger: Logger, user_name: str, host: str, workspace_id: str | None):
    from vectice.utils.logging_utils import CONNECTION_LOGGING

    if workspace_id:
        logging_output = f"For quick access to your default workspace in the Vectice web app, visit:\n{host}/browse/workspace/{workspace_id}"
        _logger.info(CONNECTION_LOGGING.format(user=user_name, logging_output=logging_output))
        return
    logging_output = f"For quick access to the list of workspaces in the Vectice web app, visit:\n{host}/workspaces"
    _logger.info(CONNECTION_LOGGING.format(user=user_name, logging_output=logging_output))


def _register_dataset_logging(
    step: Step, data: dict, value: Any, attachments, _logger: Logger, step_artifacts: Any | None = None
):
    iteration_id = step._iteration.id
    url = step._client.auth.api_base_url
    hyper_link = f"{url}/browse/iteration/{iteration_id}"

    # check if exists
    existing_logging = existing_dataset_logger(data, value.name, _logger)
    # check if attached to step
    check_artifacts = step_artifacts if step_artifacts else step.artifacts
    match = next(filter(lambda x: x.dataset_version_id == data["datasetVersion"]["vecticeId"], check_artifacts), None)
    attachments_output = None
    if attachments:
        attachments_output = ", ".join([attach["fileName"] for attach in attachments])
    logging_output = None
    if existing_logging and match:
        logging_output = dedent(
            f"""
                                Existing Dataset: {value.name!r} and Version: {data['datasetVersion']['name']!r} already linked to Step: {step.name}
                                Attachments: {attachments_output}
                                Link to Step: {hyper_link}
                                """
        ).lstrip()
    if existing_logging and not match:
        logging_output = dedent(
            f"""
                    New Version: {data['datasetVersion']['name']!r} of Dataset: {value.name!r} added to Step: {step.name}
                    Attachments: {attachments_output}
                    Link to Step: {hyper_link}
                    """
        ).lstrip()
    if not existing_logging and not match:
        logging_output = dedent(
            f"""
                                    New Dataset: {value.name!r} Version: {data['datasetVersion']['name']!r} added to Step: {step.name}
                                    Attachments: {attachments_output}
                                    Link to Step: {hyper_link}
                                    """
        ).lstrip()
    if logging_output:
        _logger.info(logging_output)
    else:
        _logger.debug("Logging failed for register dataset at step, check _register_dataset_logging.")


def _comment_or_image_logging(step: Step, _logger: Logger, filename: str | None = None):
    iteration_id = step._iteration.id
    url = step._client.auth.api_base_url
    hyper_link = f"{url}/browse/iteration/{iteration_id}"
    if filename:
        artifact_reference = f"Image: {filename!r}"
    else:
        artifact_reference = "Comment"
    logging_output = dedent(
        f"""
        Added {artifact_reference} to Step: {step.name}

        Link to Step: {hyper_link}
        """
    ).lstrip()

    _logger.info(logging_output)


def _register_model_logging(
    step: Step, data: dict, value: Any, step_name: str, attachments, _logger: Logger, step_artifacts: Any | None = None
):
    iteration_id = step._iteration.id
    url = step._client.auth.api_base_url
    hyper_link = f"{url}/browse/iteration/{iteration_id}"

    # check if exists
    existing_logging = existing_model_logger(data, value.name, _logger)
    # check if attached to step already
    check_artifacts = step_artifacts if step_artifacts else step.artifacts
    match = next(filter(lambda x: x.model_version_id == data["modelVersion"]["vecticeId"], check_artifacts), None)
    attachments_output = None
    if attachments:
        attachments_output = ", ".join([attach["fileName"] for attach in attachments])
    logging_output = None
    if existing_logging and match:
        logging_output = dedent(
            f"""
                                Existing Model: {value.name!r} and Version: {data['modelVersion']['name']!r} already linked to Step: {step_name}
                                Attachments: {attachments_output}
                                Link to Step: {hyper_link}
                                """
        ).lstrip()
    if existing_logging and not match:
        logging_output = dedent(
            f"""
                    New Version: {data['modelVersion']['name']!r} of Model: {value.name!r} added to Step: {step_name}
                    Attachments: {attachments_output}
                    Link to Step: {hyper_link}
                    """
        ).lstrip()
    if not existing_logging and not match:
        logging_output = dedent(
            f"""
                                    New Model: {value.name!r} Version: {data['modelVersion']['name']!r} added to Step: {step_name}
                                    Attachments: {attachments_output}
                                    Link to Step: {hyper_link}
                                    """
        ).lstrip()

    if logging_output:
        _logger.info(logging_output)
    else:
        _logger.debug("Logging failed for register model at step, check _register_model_logging.")



from __future__ import annotations

from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser
from vectice.api.json.model_representation import ModelRepresentationOutput
from vectice.api.json.model_version_representation import ModelVersionRepresentationOutput

if TYPE_CHECKING:
    from vectice.api.json.model_register import ModelRegisterInput, ModelRegisterOutput

_RETURNS = """
            modelVersion{
                          vecticeId
                          name
                          version
                          description
                          algorithmName
                          framework
                          modelId
                          model {
                            name
                            projectId
                          }
            }
            useExistingModel
            __typename
            """

_RETURNS_MODEL = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            type
            lastVersion {
                vecticeId
                createdDate
                updatedDate
                name
                description
                status
                algorithmName
                framework
                __typename
            }
            project {
                vecticeId
            }
            __typename
"""

_RETURNS_MODEL_VERSION = """
            vecticeId
            createdDate
            updatedDate
            name
            description
            status
            algorithmName
            framework
            model {
                project {
                    vecticeId
                }
            }
            __typename
"""


class GqlModelApi(GqlApi):
    def get_model(self, id: str) -> ModelRepresentationOutput:
        variable_types = "$modelId:VecticeId!"
        kw = "modelId:$modelId"
        variables = {"modelId": id}
        query = GqlApi.build_query(
            gql_query="getModelById",
            variable_types=variable_types,
            returns=_RETURNS_MODEL,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelRepresentationOutput = Parser().parse_item(response["getModelById"])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model", id)

    def get_model_version(self, id: str) -> ModelVersionRepresentationOutput:
        variable_types = "$modelVersionId:VecticeId!"
        kw = "modelVersionId:$modelVersionId"
        variables = {"modelVersionId": id}
        query = GqlApi.build_query(
            gql_query="getModelVersion",
            variable_types=variable_types,
            returns=_RETURNS_MODEL_VERSION,
            keyword_arguments=kw,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelVersionRepresentationOutput = Parser().parse_item(response["getModelVersion"])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model_version", id)

    def register_model(
        self,
        data: ModelRegisterInput,
        project_id: str,
        phase_id: str | None = None,
        iteration_id: str | None = None,
    ) -> ModelRegisterOutput:
        variables = {"projectId": project_id, "data": data}
        kw = "projectId:$projectId,data:$data"
        variable_types = "$projectId:VecticeId!,$data:ModelRegisterInput!"
        if phase_id:
            variable_types += ",$phaseId:VecticeId!"
            kw += ",phaseId:$phaseId"
            variables["phaseId"] = phase_id
        if iteration_id:
            variable_types += ",$iterationId:VecticeId!"
            kw += ",iterationId:$iterationId"
            variables["iterationId"] = iteration_id
        query_name = "registerModel"
        query = GqlApi.build_query(
            gql_query=query_name,
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            model_output: ModelRegisterOutput = Parser().parse_item(response[query_name])
            return model_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "model", "register model")



from io import IOBase

from PIL.Image import Image

from vectice.utils.common_utils import _check_image_path


def is_image(value):
    return is_pil_image(value) or is_binary(value) or is_existing_image_path(value)


def is_pil_image(value):
    return value if isinstance(value, Image) else None


def is_binary(value):
    return value if isinstance(value, IOBase) else None


def is_existing_image_path(value):
    return value if isinstance(value, str) and _check_image_path(value) else None



from __future__ import annotations

from vectice.api.json.workspace import WorkspaceOutput


class ProjectInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str:
        return str(self["description"])


class ProjectOutput(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "workspace" in self:
            self._workspace: WorkspaceOutput = WorkspaceOutput(**self["workspace"])

    @property
    def id(self) -> str:
        return str(self["vecticeId"])

    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def description(self) -> str | None:
        if "description" in self and self["description"] is not None:
            return str(self["description"])
        else:
            return None

    @property
    def workspace(self) -> WorkspaceOutput:
        return self._workspace



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from vectice.api.json.dataset_version import DatasetVersionOutput
    from vectice.models.property import Property
    from vectice.models.resource.metadata import FilesMetadata, MetadataDB


class DatasetRegisterOutput(dict):
    @property
    def dataset_version(self) -> DatasetVersionOutput:
        from vectice.api.json.dataset_version import DatasetVersionOutput

        return DatasetVersionOutput(**self["datasetVersion"])

    @property
    def use_existing_version(self) -> bool:
        return bool(self["useExistingVersion"])

    @property
    def use_existing_dataset(self) -> bool:
        return bool(self["useExistingDataset"])

    # TODO: complete the jobrun property
    @property
    def job_run(self) -> str:
        return str(self["jobRun"])


class DatasetRegisterInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def data_sources(self) -> list[dict]:
        return list(self["datasetSources"])

    @property
    def derived_from(self) -> list[int]:
        return list(self["inputs"])

    @property
    def properties(self) -> Property:
        from vectice.models.property import Property

        return Property(**self["properties"])


class DatasetSourceInput(dict):
    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def usage(self) -> str:
        return str(self["usage"])

    @property
    def origin(self) -> str:
        return str(self["origin"])

    @property
    def size(self) -> int:
        return int(self["size"])

    @property
    def dbs(self) -> list[MetadataDB]:
        from vectice.models.resource.metadata import MetadataDB

        return [MetadataDB(**item) for item in self["dbs"]]

    @property
    def files(self) -> FilesMetadata:
        from vectice.models.resource.metadata import FilesMetadata

        return FilesMetadata(**self["files"])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from gql import gql
from gql.transport.exceptions import TransportQueryError

from vectice.api.gql_api import GqlApi, Parser

if TYPE_CHECKING:
    from vectice.api.json.code import CodeInput, CodeOutput


_logger = logging.getLogger(__name__)


_RETURNS = """
            createdDate
            updatedDate
            deletedDate
            version
            id
            name
            description
            uri
            projectId
            authorId
            lastUpdatedById
            __typename
            """


class GqlCodeApi(GqlApi):
    def create_code(self, project_id: str, code: CodeInput) -> CodeOutput:
        variable_types = "$projectId:VecticeId!,$code:CodeCreateInput!"
        kw = "projectId:$projectId,code:$code"
        variables = {"projectId": project_id, "code": code}

        query = GqlApi.build_query(
            gql_query="createCode",
            variable_types=variable_types,
            returns=_RETURNS,
            keyword_arguments=kw,
            query=False,
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            dataset_output: CodeOutput = Parser().parse_item(response["createCode"])
            return dataset_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "code", "createCode")

    def get_code(self, code: str | int, project_id: str | None = None) -> CodeOutput:
        if isinstance(code, int):
            gql_query = "getCode"
            variable_types = "$codeId:Float!"
            variables = {"codeId": code}
            kw = "codeId:$codeId"
        elif isinstance(code, str) and project_id:
            gql_query = "getCodeByName"
            variable_types = "$name:String!,$projectId:VecticeId!"
            variables = {"name": code, "projectId": project_id}  # type: ignore[dict-item]
            kw = "name:$name,projectId:$projectId"
        else:
            raise ValueError("Missing parameters: string and parent id required.")
        query = GqlApi.build_query(
            gql_query=gql_query, variable_types=variable_types, returns=_RETURNS, keyword_arguments=kw, query=True
        )
        query_built = gql(query)
        try:
            response = self.execute(query_built, variables)
            phase_output: CodeOutput = Parser().parse_item(response[gql_query])
            return phase_output
        except TransportQueryError as e:
            self._error_handler.handle_post_gql_error(e, "getCode", code)



from __future__ import annotations

import logging
import re
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

from rich.table import Table

from vectice.api.http_error_handlers import (
    InvalidIdError,
    LastIterationNotWritableError,
    MultipleActiveIterationsError,
)
from vectice.api.json.iteration import IterationOutput, IterationStatus
from vectice.api.json.phase import PhaseStatus
from vectice.models.iteration import Iteration
from vectice.utils.common_utils import _temp_print
from vectice.utils.logging_utils import get_iteration_status
from vectice.utils.vectice_ids_regex import ITERATION_VID_REG

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Project, Workspace


_logger = logging.getLogger(__name__)


class Phase:
    """Represent a Vectice phase.

    Phases reflect the real-life phases of the project lifecycle
    (i.e., Business Understanding, Data Preparation, Modeling,
    Deployment, etc.).  The Vectice admin creates the phases of a
    project.

    Phases let you document the goals, assets, and outcomes along with
    the status to organize the project, enforce best practices, allow
    consistency, and capture knowledge.

    Phases contain definitions of steps that are performed
    by data-scientists in order to complete iterations.

    ```tree
    phase 1
        step definition 1
        step definition 2
        step definition 3
    ```

    To get a project's phase:

    ```python
    my_phase = my_project.phase("Business Understanding")
    ```

    Iterations can then be created for this phase,
    to complete the phase steps:

    ```python
    my_origin_dataset = ...
    my_iteration = my_phase.create_iteration()
    my_iteration.step_origin_dataset = my_origin_dataset
    ```

    NOTE: **Phases and Steps Definitions are created in the Vectice App,
    Iterations are created from the Vectice Python API.**

    See the documentation of [Iterations][vectice.models.Iteration]
    for more information about iterations.
    """

    __slots__: ClassVar[list[str]] = [
        "_id",
        "_project",
        "_name",
        "_index",
        "_status",
        "_client",
        "_current_iteration",
        "_pointers",
    ]

    def __init__(
        self,
        id: str,
        project: Project,
        name: str,
        index: int,
        status: PhaseStatus = PhaseStatus.NotStarted,
    ):
        """Initialize a phase.

        Vectice users shouldn't need to instantiate Phases manually,
        but here are the phase parameters.

        Parameters:
            id: The phase identifier.
            project: The project to which the phase belongs.
            name: The name of the phase.
            index: The index of the phase.
            status: The status of the phase.
        """
        self._id = id
        self._project = project
        self._name = name
        self._index = index
        self._status = status
        self._client: Client = self._project._client
        self._current_iteration: Iteration | None = None

    def __repr__(self):
        return f"Phase (name='{self.name}', id={self.id}, status='{self.status.name}')"

    def __eq__(self, other: object):
        if not isinstance(other, Phase):
            return NotImplemented
        return self.id == other.id

    @property
    def id(self) -> str:
        """The phase's id.

        Returns:
            The phase's id.
        """
        return self._id

    @id.setter
    def id(self, phase_id: str):
        """Set the phase's id.

        Parameters:
            phase_id: The phase id to set.
        """
        self._id = phase_id

    @property
    def name(self) -> str:
        """The phase's name.

        Returns:
            The phase's name.
        """
        return self._name

    @property
    def index(self) -> int:
        """The phase's index.

        Returns:
            The phase's index.
        """
        return self._index

    @property
    def status(self) -> PhaseStatus:
        """The phase's status.

        Returns:
            The phase's status.
        """
        return self._status

    @property
    def properties(self) -> dict:
        """The phase's name, id, and index.

        Returns:
            A dictionary containing the `name`, `id`, and `index` items.
        """
        return {"name": self.name, "id": self.id, "index": self.index}

    def list_iterations(self, only_mine: bool = False, statuses: list[IterationStatus] | None = None) -> None:
        """Prints a list of iterations belonging to the phase in a tabular format, limited to the last 10 items.

        Parameters:
            only_mine: Display only the iterations where the user is the owner.
            statuses: Filter iterations on specified statuses.

        Returns:
            None

        """
        iteration_outputs = self._client.list_iterations(self.id, only_mine, statuses)
        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Id", justify="left", no_wrap=True, min_width=3, max_width=20)
        rich_table.add_column("Index", justify="left", no_wrap=True, min_width=5, max_width=10)
        rich_table.add_column("Status", justify="left", no_wrap=True, min_width=3, max_width=15)
        rich_table.add_column("Owner", justify="left", no_wrap=True, min_width=5, max_width=15)

        for iteration in iteration_outputs.list:
            rich_table.add_row(
                str(iteration.id),
                str(iteration.index),
                get_iteration_status(iteration.status, iteration.starred),
                iteration.ownername,
            )

        iteration_statuses_log = (
            " | ".join(list(map(lambda status: get_iteration_status(status), statuses)))
            if statuses is not None and len(statuses) > 0
            else None
        )
        status_log = f"with status [{iteration_statuses_log}] " if iteration_statuses_log is not None else ""
        description = f"""There are {iteration_outputs.total} iterations {status_log}in the phase {self.name!r} and a maximum of 10 iterations are displayed in the table below:"""
        link = dedent(
            f"""
        # For quick access to the list of iterations in the Vectice web app, visit:
        # {self._client.auth._API_BASE_URL}/phase/{self.id}/iterations?w={self.workspace.id}"""
        ).lstrip()
        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(link)

    def create_or_get_current_iteration(self) -> Iteration | None:
        """Get or create an iteration.

        If your last updated iteration is writable (Not Started or In Progress), Vectice will return it.
        Otherwise, Vectice will create a new one and return it.
        If multiple writable iterations are found, Vectice will print a list of the iterations to complete or cancel.

        Returns:
            An iteration or None if Vectice could not determine which iteration to retrieve.

        Raises:
            VecticeException: When attempting to create an iteration but there isn't any step inside the phase.
        """
        try:
            iteration_output = self._client.get_last_iteration(self.id)
        except (LastIterationNotWritableError, MultipleActiveIterationsError) as e:
            _logger.warning(str(e.value))
            self.list_iterations(True, [IterationStatus.NotStarted, IterationStatus.InProgress])
            return None

        return self._build_iteration_from_output_and_log(iteration_output)

    def iteration(self, index: int | str) -> Iteration:
        """Get an iteration.

        Fetch and return an iteration by index or id.

        Parameters:
            index: The index or id of an existing iteration.

        Returns:
            An iteration.

        Raises:
            InvalidIdError: When index is a string and not matching 'ITR-[int]'
            IterationIdError: Iteration with specified id does not exist.
            IterationIndexError: Iteration with specified index does not exist.
        """
        if isinstance(index, str):
            if re.search(ITERATION_VID_REG, index):
                iteration_output = self._client.get_iteration_by_id(index)
            else:
                raise InvalidIdError("iteration", index)
        else:
            iteration_output = self._client.get_iteration_by_index(self.id, index)

        return self._build_iteration_from_output_and_log(iteration_output)

    def _build_iteration_from_output_and_log(self, iteration_output: IterationOutput) -> Iteration:
        iteration = Iteration(
            id=iteration_output.id,
            index=iteration_output.index,
            phase=self,
            status=iteration_output.status,
        )
        base_log = dedent(f"Iteration number {iteration.index!r} successfully retrieved.")
        if iteration.status in [IterationStatus.Completed, IterationStatus.Abandoned]:
            base_log += dedent(
                f"""
                WARN: Iteration is {iteration.status}."""
            )

        logging_output = dedent(
            f"""
                {base_log}

                For quick access to the Iteration in the Vectice web app, visit:
                {self._client.auth._API_BASE_URL}/browse/iteration/{iteration.id}"""
        ).lstrip()
        _logger.info(logging_output)
        self._current_iteration = iteration
        return iteration

    def create_iteration(self) -> Iteration:
        """Create a new iteration.

        Create and return an iteration.

        Returns:
            An iteration.
        """
        iteration_output = self._client.create_iteration(self.id)
        logging_output = dedent(
            f"""
        New Iteration number {iteration_output.index!r} created.

        For quick access to the Iteration in the Vectice web app, visit:
        {self._client.auth._API_BASE_URL}/browse/iteration/{iteration_output.id}"""
        ).lstrip()
        _logger.info(logging_output)

        iteration_object = Iteration(
            id=iteration_output.id,
            index=iteration_output.index,
            phase=self,
            status=iteration_output.status,
        )
        self._current_iteration = iteration_object
        return iteration_object

    @property
    def connection(self) -> Connection:
        """The connection to which this phase belongs.

        Returns:
            The connection to which this phase belongs.
        """
        return self._project.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this phase belongs.

        Returns:
            The workspace to which this phase belongs.
        """
        return self._project.workspace

    @property
    def project(self) -> Project:
        """The project to which this phase belongs.

        Returns:
            The project to which this phase belongs.
        """
        return self._project



from __future__ import annotations

import logging
from contextlib import suppress
from textwrap import dedent
from typing import TYPE_CHECKING, ClassVar

from rich.table import Table

from vectice.api.http_error_handlers import NoStepsInPhaseError, VecticeException
from vectice.api.json.iteration import (
    IterationInput,
    IterationStatus,
)
from vectice.utils.common_utils import _check_read_only, _get_step_type, _temp_print
from vectice.utils.logging_utils import get_iteration_status

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.api import Client
    from vectice.models import Phase, Project, Step, Workspace
    from vectice.models.model import Model

_logger = logging.getLogger(__name__)


class Iteration:
    """Represent a Vectice iteration.

    Iterations reflect the model development and test cycles completed
    by data scientists until a fully functional algorithm is ready for
    deployment.  Each iteration contains the sequence of steps defined
    at the Phase and acts as a guardrail for data scientists to
    provide their updates.

    Typical usage example:

    ```python
    my_iteration = my_phase.create_iteration()

    my_iteration.step_cleaning = my_dataset
    # you can append assets of the same types (datasets/models) with +=
    my_iteration.step_cleaning += my_other_dataset

    my_iteration.step_model = my_model
    my_iteration.step_model += my_other_model
    ```

    If steps are added to a phase after iterations have been created
    and completed, these steps won't appear in these iterations.

    ```tree
    iteration 1
        step 1
        step 2
        step 3
    ```

    NOTE: **Phases and Steps Definitions are created in the Vectice App,
    Iterations are created from the Vectice Python API.**

    To create a new iteration:

    ```python
    my_iteration = my_phase.create_iteration()
    ```
    """

    __slots__: ClassVar[list[str]] = [
        "_id",
        "_index",
        "_phase",
        "_status",
        "_client",
        "_model",
        "_current_step",
        "_steps",
        "__dict__",
    ]

    def __init__(
        self,
        id: str,
        index: int,
        phase: Phase,
        status: IterationStatus = IterationStatus.NotStarted,
    ):
        """Initialize an iteration.

        Vectice users shouldn't need to instantiate Iterations manually,
        but here are the iteration parameters.

        Parameters:
            id: The iteration identifier.
            index: The index of the iteration.
            phase: The project to which the iteration belongs.
            status: The status of the iteration.
        """
        self.__dict__ = {}
        self._id = id
        self._index = index
        self._phase = phase
        self._status = status
        self._client: Client = self._phase._client
        self._model: Model | None = None
        self._current_step: Step | None = None
        self._steps = self._populate_steps()

    def __repr__(self):
        steps = len(self.steps)
        return f"Iteration (index={self._index}, status={self._status}, No. of steps={steps})"

    def __eq__(self, other: object):
        if not isinstance(other, Iteration):
            return NotImplemented
        return self.id == other.id

    def _populate_steps(self):
        with suppress(NoStepsInPhaseError):
            check = {step.slug: step for step in self.steps}
            return check
        return {}

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]
        if item in super().__getattribute__("_steps"):
            step_item = self._steps[item]
            step = self._client.get_step_by_name(step_item.name, self.id)
            step_object = _get_step_type(
                id=step.id,
                iteration=self,
                name=step.name,
                index=step.index,
                slug=step.slug,
                description=step.description,
                artifacts=step.artifacts,
            )
            return step_object
        raise AttributeError(f"The attribute '{item}' does not exist.")

    def __setattr__(self, attr_name, attr_value):
        if hasattr(self, "_steps") and attr_name in super().__getattribute__("_steps"):
            if self._status in {IterationStatus.Abandoned, IterationStatus.Completed}:
                raise VecticeException(f"The Iteration is {self.status} and is read-only!")
            self._steps[attr_name] = self._steps[attr_name]._step_factory_and_update(value=attr_value)
        elif hasattr(self, "__dict__") and attr_name not in self.__slots__:
            raise AttributeError(f"The attribute '{attr_name}' does not exist.")
        else:
            super().__setattr__(attr_name, attr_value)

    @property
    def id(self) -> str:
        """The iteration's identifier.

        Returns:
            The iteration's identifier.
        """
        return self._id

    @id.setter
    def id(self, iteration_id: str):
        """Set the iteration's identifier.

        Parameters:
            iteration_id: The identifier.
        """
        _check_read_only(self)
        self._id = iteration_id

    @property
    def index(self) -> int:
        """The iteration's index.

        Returns:
            The iteration's index.
        """
        return self._index

    @property
    def status(self) -> str:
        """The iteration's status.

        Returns:
            The iteration's status.
        """
        return get_iteration_status(self._status)

    @property
    def completed(self) -> bool:
        """Whether this iteration is completed.

        Returns:
            Whether the iteration is completed.
        """
        return self._status is IterationStatus.Completed

    @property
    def properties(self) -> dict:
        """The iteration's identifier and index.

        Returns:
            A dictionary containing the `id` and `index` items.
        """
        return {"id": self.id, "index": self.index}

    @property
    def step_names(self) -> list[str]:
        """The names of the steps required in this iteration.

        Returns:
            The steps names.
        """
        return [step.name for step in self.steps]

    @property
    def steps(self) -> list[Step]:
        """The steps required in this iteration.

        Returns:
            The steps required in this iteration.
        """
        steps_output = self._client.list_steps(self.id)
        return [
            _get_step_type(
                id=item.id,
                iteration=self,
                name=item.name,
                index=item.index,
                slug=item.slug,
                description=item.description,
                artifacts=item.artifacts,
            )
            for item in steps_output.list
        ]

    def list_steps(self) -> None:
        """Prints a list of steps belonging to the iteration in a tabular format, limited to the first 10 steps. A link is provided to view the remaining steps.

        Returns:
            None
        """
        steps_output = self._client.list_steps(self.id, populate=False)

        rich_table = Table(expand=True, show_edge=False)

        rich_table.add_column("Shortcut", justify="left", no_wrap=True, min_width=5, max_width=20)
        rich_table.add_column("Registered items (count)", justify="left", no_wrap=True, min_width=5, max_width=15)
        rich_table.add_column("Description", justify="left", no_wrap=True, min_width=5, max_width=35)

        for step in steps_output.list:
            rich_table.add_row(step.slug, str(step.artifacts_count), step.description)
        description = f"""There are {steps_output.total} steps in Iteration '{self.index!r}' and a maximum of 10 steps are displayed in the table below:"""
        tips = dedent(
            """
        To access a specific step, use the step shortcut \033[1iiteration\033[0m.step_my_step_name
        The step reference is referred to as shortcut"""
        ).lstrip()
        link = dedent(
            f"""
        For quick access to the list of steps in the Vectice web app, visit:
        {self._client.auth._API_BASE_URL}/browse/iteration/{self.id}"""
        ).lstrip()

        _temp_print(description)
        _temp_print(table=rich_table)
        _temp_print(tips)
        _temp_print(link)

    def cancel(self) -> None:
        """Cancel the iteration by abandoning all unfinished steps."""
        iteration_input = IterationInput(status=IterationStatus.Abandoned.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Abandoned
        _logger.info(f"Iteration with index {self.index} cancelled.")

    def complete(self) -> None:
        """Mark the iteration as completed."""
        if self._status is IterationStatus.Abandoned:
            raise VecticeException("The iteration is cancelled and cannot be completed.")
        iteration_input = IterationInput(status=IterationStatus.Completed.name)
        self._client.update_iteration(self.id, iteration_input)
        self._status = IterationStatus.Completed
        logging_output = dedent(
            f"""
                        Iteration with index {self.index} completed.

                        For quick access to the Iteration in the Vectice web app, visit:
                        {self._client.auth._API_BASE_URL}/browse/iteration/{self.id}"""
        ).lstrip()
        _logger.info(logging_output)

    def delete(self) -> None:
        """Permanently deletes the iteration."""
        self._client.delete_iteration(self.id)
        _logger.info(f"Iteration with index {self.index} was deleted.")

    @property
    def connection(self) -> Connection:
        """The connection to which this iteration belongs.

        Returns:
            The connection to which this iteration belongs.
        """
        return self._phase.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this iteration belongs.

        Returns:
            The workspace to which this iteration belongs.
        """
        return self._phase.workspace

    @property
    def project(self) -> Project:
        """The project to which this iteration belongs.

        Returns:
            The project to which this iteration belongs.
        """
        return self._phase.project

    @property
    def phase(self) -> Phase:
        """The phase to which this iteration belongs.

        Returns:
            The phase to which this iteration belongs.
        """
        return self._phase



from __future__ import annotations

import logging
from typing import Dict

from pandas.core.frame import DataFrame as PandasDF
from pyspark.pandas.frame import DataFrame as PysparkPandasDF
from pyspark.sql.dataframe import DataFrame as SparkDF

from vectice.models.resource.metadata.base import MetadataSettings
from vectice.models.resource.metadata.column_metadata import Column
from vectice.models.resource.metadata.dataframe_config import DataFrameType
from vectice.models.resource.metadata.df_wrapper_pandas import PandasDFWrapper
from vectice.models.resource.metadata.df_wrapper_pyspark_pandas import PysparkPandasDFWrapper
from vectice.models.resource.metadata.df_wrapper_resource import DataFrameWrapper
from vectice.models.resource.metadata.df_wrapper_spark_df import SparkDFWrapper

_logger = logging.getLogger(__name__)


class Source:
    def __init__(
        self,
        name: str,
        size: int | None = None,
        columns: list[Column] | None = None,
        updated_date: str | None = None,
        created_date: str | None = None,
        dataframe: DataFrameType | None = None,
    ):
        """Initialize a MetadataDB instance.

        Parameters:
            name: The name of the source.
            size: The size of the source.
            columns: The columns that compose the source.
            updated_date: The date of last update of the source.
            created_date: The date of last update of the source.
            dataframe (Optional): A dataframe allowing vectice to optionally compute more metadata about this resource such as columns stats, size, rows number and column numbers. (Support Pandas)
        """
        self.name = name
        self.size = size
        self.columns = columns
        self.created_date = created_date
        self.updated_date = updated_date
        self._dataframe = dataframe
        self._wrapper: PandasDFWrapper | PysparkPandasDFWrapper | SparkDFWrapper | None = None
        self._settings = MetadataSettings()

        if isinstance(self._dataframe, PandasDF):
            self._wrapper = PandasDFWrapper(self._dataframe)
        elif isinstance(self._dataframe, SparkDF):
            self._wrapper = SparkDFWrapper(self._dataframe)
        elif isinstance(self._dataframe, PysparkPandasDF):
            _logger.warning(
                "WARNING: Pandas on spark is not supported yet, pass a Pandas or a Spark DataFrame to get statistics."
            )
        elif isinstance(self._dataframe, DataFrameWrapper):
            _logger.warning(
                "WARNING: Custom wrappers are not supported yet, pass a Pandas or a Spark DataFrame to get statistics."
            )

    def set_settings(self, settings: MetadataSettings):
        self._settings = settings

    def asdict(self) -> dict:
        size_info: Dict[str, int | float | None] = {"rowsNumber": None, "columnsNumber": None}
        skipped_statistics = False
        if self._wrapper is not None:
            df_info = self._wrapper.get_size()
            minimum_rows_for_statistics = self._settings.minimum_rows_for_statistics
            if df_info.rows < minimum_rows_for_statistics:
                skipped_statistics = True
                _logger.warning(
                    f"Statistics are not captured if numbers of rows are below {minimum_rows_for_statistics}, to keep the data anonymous."
                )
            size_info = df_info.asdict()

        columns_list: list[Column] = self.columns if self.columns is not None else []
        columns_list = (
            self._wrapper.capture_columns(minimum_rows_for_statistics) if self._wrapper is not None else columns_list
        )
        columns_list_dict = [col.asdict() for col in columns_list]

        return {
            **size_info,
            "size": self.size,
            "name": self.name,
            "updatedDate": self.updated_date,
            "createdDate": self.created_date,
            "columns": columns_list_dict,
            "skippedStatistics": skipped_statistics,
        }



from __future__ import annotations

import re
from typing import TYPE_CHECKING, Any, NoReturn

from vectice.utils.vectice_ids_regex import ITERATION_VID_REG, PHASE_VID_REG, PROJECT_VID_REG, WORKSPACE_VID_REG

if TYPE_CHECKING:
    from gql.transport.exceptions import TransportQueryError

    from vectice.api.http_error import HttpError

BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE = "Bad or missing credentials"


class InvalidIdError(ValueError):
    """When an incorrect value type is passed at the client level."""

    def __init__(self, reference_type: str, value: Any) -> None:
        valid_id = "a valid id"
        wsp = "WSP-[int]"
        prj = "PRJ-[int]"
        pha = "PHA-[int]"
        itr = "ITR-[int]"
        dts = "DTS-[int]"
        dsv = "DTV-[int]"
        mod = "MDL-[int]"
        mdv = "MDV-[int]"
        if reference_type == "workspace":
            valid_id = wsp
        elif reference_type == "project":
            valid_id = prj
        elif reference_type == "phase":
            valid_id = pha
        elif reference_type == "iteration":
            valid_id = itr
        elif reference_type == "dataset":
            valid_id = dts
        elif reference_type == "dataset_version":
            valid_id = dsv
        elif reference_type == "model":
            valid_id = mod
        elif reference_type == "model_version":
            valid_id = mdv
        elif reference_type == "asset":
            refs = {", ".join([wsp, prj, pha, itr, dts, dsv, mod, mdv])}
            valid_id = f"one of ({refs})"

        super().__init__(
            f"The {reference_type} reference is invalid. Please check the provided value. "
            + f"It should be {valid_id} "
            + f"Provided value is {value}"
        )


class InvalidReferenceError(ValueError):
    """When an incorrect value type is passed at the client level."""

    def __init__(self, reference_type: str, value: Any) -> None:
        super().__init__(
            f"The {reference_type} reference is invalid."
            + "Please check the provided value as it should be a string or a number."
            + f"Provided value is {value} ({type(value)})"
        )


class MissingReferenceError(ValueError):
    """When a value is missing at the client level."""

    def __init__(self, reference_type: str, parent_reference_type: str | None = None) -> None:
        if parent_reference_type is not None:
            super().__init__(f"The {parent_reference_type} reference is required if the {reference_type} name is given")
        else:
            super().__init__(f"The {reference_type} reference is required")


class ClientErrorHandler:
    def _graphql_error_formatter(self, error: TransportQueryError) -> Exception:
        error_field = error.errors[0] if error.errors else {}
        error_message = "An unknown error occured, please contact support."
        if "extensions" in error_field and "exception" in error_field["extensions"]:
            exception = error_field["extensions"]["exception"]
            if "stacktrace" in exception:
                # try to format the message
                error_code = 500
                message = exception["stacktrace"]
                if message and len(message) >= 1 and isinstance(message[0], str):
                    message = message[0].split(":")
                    if len(message) > 1:
                        error_code = message[0]
                        error_message = message[1]
                raise VecticeException(f"{error_code}: {error_message}")
        elif "message" in error_field:
            error_message = error_field["message"]

        # just dump the stringified error
        raise VecticeException(repr(error_message))

    def handle_code(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        if e.code == 404:
            raise BadReferenceFactory.get_reference(reference_type, reference)
        elif e.code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif e.code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        raise RuntimeError(f"Can not access {reference_type}: {e.reason}")

    def handle_get_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_put_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_delete_http_error(self, e: HttpError, reference_type: str, reference: str | int) -> NoReturn:
        self.handle_code(e, reference_type, reference)

    def handle_post_http_error(self, e: HttpError, reference_type: str, action: str = "create") -> NoReturn:
        if e.code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif e.code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        elif e.code == 400:
            raise RuntimeError(f"Can not {action} {reference_type}: {e.reason}")
        raise RuntimeError(f"Unexpected error: {e.reason}")

    def handle_post_gql_error(self, error: TransportQueryError, reference_type: str, reference: str | int) -> NoReturn:
        try:
            status_code = error.errors[0]["extensions"]["exception"]["status"]  # type: ignore[index]
        except KeyError:
            status_code = -1
        if status_code == 404:
            raise BadReferenceFactory.get_reference(reference_type, reference)
        elif status_code == 401:
            raise ConnectionRefusedError(BAD_OR_MISSING_CREDENTIALS_ERROR_MESSAGE)
        elif status_code == 403:
            raise PermissionError(f"Missing permissions to access this {reference_type}")
        elif status_code == 400 or status_code == 412:
            raise BadRequestFactory.get_reference(reference_type, error)
        raise self._graphql_error_formatter(error) from error


class BadRequestFactory:
    @classmethod
    def get_reference(cls, reference_type: str, error: TransportQueryError) -> Exception:
        message: str = "failed"
        key: str | None = None
        try:
            message = error.errors[0]["message"]  # type: ignore
            key = error.errors[0]["extensions"]["key"]  # type: ignore
        except KeyError:
            pass

        if key == "iteration_not_writable":
            raise LastIterationNotWritableError()
        elif key == "iteration_multiple_active":
            raise MultipleActiveIterationsError()

        raise VecticeException(f"{reference_type}: {message}")


class BadReferenceFactory:
    """Appropriate exceptions for HTTP/GQL errors."""

    @classmethod
    def get_reference(cls, reference_type: str, value: str | int) -> Exception:
        """Get an appropriate error given a reference type.

        Parameters:
            reference_type: The reference type.
            value: A value.

        Returns:
            An exception instance.
        """
        if isinstance(value, str) and reference_type == "workspace":
            if not re.search(WORKSPACE_VID_REG, value):
                return WorkspaceNameError(value)
            return WorkspaceIdError(value)
        elif isinstance(value, str) and reference_type == "project":
            if not re.search(PROJECT_VID_REG, value):
                return ProjectNameError(value)
            return ProjectIdError(value)
        elif isinstance(value, str) and reference_type == "phase":
            if not re.search(PHASE_VID_REG, value):
                return PhaseNameError(value)
            return PhaseIdError(value)
        elif reference_type == "step":
            if isinstance(value, str):
                return StepNameError(value)
            elif isinstance(value, int):
                return StepIdError(value)
        elif isinstance(value, str) and reference_type == "iteration":
            if re.search(ITERATION_VID_REG, value):
                return IterationIdError(value)
            return IterationNameError(value)
        elif reference_type == "iteration_index":
            if isinstance(value, int):
                return IterationIndexError(value)
        elif isinstance(value, str) and reference_type == "steps":
            return NoStepsInPhaseError(value)
        elif isinstance(value, str) and reference_type == "dataset":
            return DatasetIdError(value)
        elif isinstance(value, str) and reference_type == "dataset_version":
            return DatasetVersionIdError(value)
        elif isinstance(value, str) and reference_type == "model":
            return ModelIdError(value)
        elif isinstance(value, str) and reference_type == "model_version":
            return ModelVersionIdError(value)
        return RuntimeError(f"The value {value} of type {reference_type} is not valid!")


class VecticeException(Exception):  # noqa: N818
    def __init__(self, value):
        super().__init__(value)
        self.__suppress_context__ = True
        self.value = value


class VecticeBaseNameError(NameError):
    def __init__(self, value):
        super().__init__(value)
        self.__suppress_context__ = True
        self.value = value

    def __str__(self):
        divider = "=" * len(f"{self.__class__.__name__}: {self.value}")
        return f"\n{divider}\n{self.__class__.__name__}: {self.value}"


class WorkspaceNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The workspace with name '{value}' is unknown.")


class WorkspaceIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The workspace with id '{value}' is unknown.")


class ProjectNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The project with name '{value}' is unknown.")


class ProjectIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The project with id '{value}' is unknown.")


class PhaseNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The phase with name '{value}' is unknown.")


class PhaseIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The phase with id '{value}' is unknown.")


class StepNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(
            f"The step with name '{value}' is unknown. Use <your iteration>.list_steps() method to find step names."
        )


class StepIdError(VecticeBaseNameError):
    def __init__(self, value: int):
        super().__init__(
            f"The step with id '{value}' is unknown. Use <your iteration>.list_steps() method to find step ids."
        )


class IterationIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The iteration with id '{value}' is unknown.")


class IterationIndexError(VecticeBaseNameError):
    def __init__(self, value: int):
        super().__init__(f"The iteration with index '{value}' is unknown.")


class IterationNameError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The iteration with name '{value}' is unknown.")


class LastIterationNotWritableError(VecticeBaseNameError):
    def __init__(self):
        super().__init__(
            "Your last iteration is not writable, and you have a previous iteration in progress. Please specify the iteration to retrieve."
        )


class MultipleActiveIterationsError(VecticeBaseNameError):
    def __init__(self):
        super().__init__("You have multiple iterations in progress. Please specify the iteration to retrieve.")


class NoStepsInPhaseError(VecticeBaseNameError):
    def __init__(self, value: str):
        ref = f"with id '{value}'" if re.search(PHASE_VID_REG, value) else f"'{value}'"
        super().__init__(f"There are no steps in the phase {ref}. Must create steps for this phase in the UI.")


class DatasetIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The dataset with id '{value}' is unknown.")


class DatasetVersionIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The dataset version with id '{value}' is unknown.")


class ModelIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The model with id '{value}' is unknown.")


class ModelVersionIdError(VecticeBaseNameError):
    def __init__(self, value: str):
        super().__init__(f"The model version with id '{value}' is unknown.")



from __future__ import annotations

import logging
import pickle  # nosec
from io import BufferedReader, IOBase
from typing import TYPE_CHECKING, Any

from PIL.Image import Image

import vectice
from vectice.api.http_error_handlers import VecticeException
from vectice.api.json.dataset_version import DatasetVersionOutput
from vectice.api.json.iteration import (
    IterationStatus,
    IterationStepArtifact,
    IterationStepArtifactInput,
    IterationStepArtifactType,
)
from vectice.api.json.model_version import ModelVersionOutput
from vectice.api.json.step import StepType
from vectice.models.attachment_container import AttachmentContainer
from vectice.utils.common_utils import _get_image_variables
from vectice.utils.instance_helper import is_image
from vectice.utils.last_assets import _comment_or_image_logging, _register_dataset_logging, _register_model_logging

if TYPE_CHECKING:
    from vectice import Connection
    from vectice.models import Iteration, Phase, Project, Workspace
    from vectice.models.dataset import Dataset
    from vectice.models.model import Model
    from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
    from vectice.models.representation.model_version_representation import ModelVersionRepresentation
    from vectice.models.step_dataset import StepDataset
    from vectice.models.step_model import StepModel
    from vectice.models.step_number import StepNumber
    from vectice.models.step_string import StepString

_logger = logging.getLogger(__name__)

MISSING_DATASOURCE_ERROR_MESSAGE = "Cannot create modeling dataset. Missing %s data source."


class Step:
    """Model a Vectice step.

    Steps define the logical sequence of steps required to complete
    the phase along with their expected outcomes.

    Steps belong to an Iteration. The steps created under a Phase are
    Step Definitions that are then re-used in Iterations to iteratively
    complete the steps until a satisfactory result is obtained.

    ```tree
    phase 1
        step definition 1
        step definition 2
        step definition 3
    ```

    ```tree
    iteration 1 of phase 1
        step 1
        step 2
        step 3
    ```

    If steps are added to a phase after iterations have been created
    and completed, these steps won't appear in these iterations.

    NOTE: **Phases and Steps Definitions are created in the Vectice App,
    Iterations are created from the Vectice Python API.**

    To complete an iteration's step, you just need to assign a value to it.
    To access the step and assign a value, you can use the "slug" of the step:
    the slug is the name of the step, transformed to fit Python's naming rules,
    and prefixed with `step_`. For example, a step called "Clean Dataset" can
    be accessed with `my_iteration.step_clean_dataset`.

    Therefore, to complete a step:

    ```python
    my_clean_dataset = ...
    my_iteration.step_clean_dataset = my_clean_dataset
    ```

    Depending on the step, you can assign it a [`Model`][vectice.models.model.Model],
    a [`Dataset`][vectice.models.dataset.Dataset], code source, or attachments.
    """

    def __init__(
        self,
        id: int,
        iteration: Iteration,
        name: str,
        index: int,
        slug: str,
        description: str | None = None,
        artifacts: list[IterationStepArtifact] | None = None,
        step_type: StepType = StepType.Step,
    ):
        """Initialize a step.

        Vectice users shouldn't need to instantiate Steps manually,
        but here are the step parameters.

        Parameters:
            id: The step identifier.
            iteration: The iteration to which the step belongs.
            name: The name of the step.
            index: The index of the step.
            slug: The slug of the step.
            description: The description of the step.
            artifacts: The artifacts linked to the steps.
            step_type: The step type.
        """
        self._id = id
        self._iteration: Iteration = iteration
        self._name = name
        self._index = index
        self._description = description
        self._client = self._iteration._client
        self._artifacts = artifacts or []
        self._slug = slug
        self._type: StepType = step_type
        self._model: Model | None = None

        self._iteration_read_only = self._iteration._status in {IterationStatus.Completed, IterationStatus.Abandoned}
        if self._iteration_read_only:
            _logger.debug(f"Step {self.name}, iteration is {self._iteration.status} and is read-only!")

    def __repr__(self):
        return f"Step(name={self.name!r}, slug={self.slug!r}, id={self.id!r})"

    def __eq__(self, other: object):
        if not isinstance(other, Step):
            return NotImplemented
        return self.id == other.id

    def __iadd__(self, value):
        # TODO cyclic import
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation
        from vectice.models.step_dataset import StepDataset
        from vectice.models.step_image import StepImage
        from vectice.models.step_model import StepModel
        from vectice.models.step_number import StepNumber
        from vectice.models.step_string import StepString

        self._ensure_correct_project_id_from_representation_objs(value)

        if is_image(value):
            artifact_input, filename = self._create_image_artifact(value)
            artifact = IterationStepArtifact(
                entity_file_id=artifact_input.id, type=IterationStepArtifactType.EntityFile
            )
            step = StepImage(self, value)
            self._iadd_comment_or_image(value, artifact)(self, _logger, filename)
        elif isinstance(value, (int, float, str)):
            self._check_string_sanity(value)
            artifact_input = IterationStepArtifactInput(text=str(value), type="Comment")
            artifact = IterationStepArtifact(text=value, type=IterationStepArtifactType.Comment)
            step = StepString(self, value) if isinstance(value, str) else StepNumber(self, value)
            self._client.add_iteration_step_artifact(self._id, artifact_input)
            self._iadd_comment_or_image(value, artifact)(self, _logger)
        elif (
            isinstance(value, Model)
            or isinstance(value, ModelRepresentation)
            or isinstance(value, ModelVersionRepresentation)
        ):
            if isinstance(value, Model):
                code_version_id = self._get_code_version_id()
                model_data = self._client.register_model(
                    model=value,
                    project_id=self.project.id,
                    phase_id=self.phase.id,
                    iteration_id=self.iteration.id,
                    code_version_id=code_version_id,
                )
                attachments_output = self._set_model_attachments(value, model_data.model_version)
                model_artifact = IterationStepArtifact(
                    modelVersionId=model_data["modelVersion"]["vecticeId"], type="ModelVersion"
                )
            elif isinstance(value, ModelRepresentation):
                model_data = {
                    "useExistingModel": True,
                    "modelVersion": {"vecticeId": value.version.id, "name": value.version.name},
                }
                model_artifact = IterationStepArtifact(
                    modelVersionId=value.version.id,
                    type=IterationStepArtifactType.ModelVersion.name,
                )
                attachments_output = self._get_version_attachments(value.version)

            elif isinstance(value, ModelVersionRepresentation):
                model_data = {
                    "useExistingModel": True,
                    "modelVersion": {"vecticeId": value.id, "name": value.name},
                }
                model_artifact = IterationStepArtifact(
                    modelVersionId=value.id,
                    type=IterationStepArtifactType.ModelVersion.name,
                )
                attachments_output = self._get_version_attachments(value)

            attachments = (
                [
                    IterationStepArtifact(entityFileId=attach["fileId"], type="EntityFile")
                    for attach in attachments_output
                ]
                if attachments_output
                else []
            )
            self.artifacts.extend([model_artifact, *attachments])
            copy_artifacts = self.artifacts.copy()
            self._keep_artifacts_and_update_step(StepType.StepModel, value)
            step = StepModel(self, model_artifact, value)
            _register_model_logging(self, model_data, value, self.name, attachments_output, _logger, copy_artifacts)
        elif (
            isinstance(value, Dataset)
            or isinstance(value, DatasetRepresentation)
            or isinstance(value, DatasetVersionRepresentation)
        ):
            if isinstance(value, Dataset):
                code_version_id = self._get_code_version_id()
                dataset_output = self._client.register_dataset_from_source(
                    dataset=value,
                    project_id=self.project.id,
                    phase_id=self.phase.id,
                    iteration_id=self.iteration.id,
                    code_version_id=code_version_id,
                )
                dataset_artifact = IterationStepArtifact(
                    datasetVersionId=dataset_output["datasetVersion"]["vecticeId"],
                    type=IterationStepArtifactType.DataSetVersion.name,
                )
                attachments_output = self._set_dataset_attachments(value, dataset_output.dataset_version)
            elif isinstance(value, DatasetRepresentation):
                dataset_output = {
                    "useExistingDataset": True,
                    "useExistingVersion": True,
                    "datasetVersion": {"vecticeId": value.version.id, "name": value.version.name},
                }
                dataset_artifact = IterationStepArtifact(
                    datasetVersionId=value.version.id,
                    type=IterationStepArtifactType.DataSetVersion.name,
                )
                attachments_output = self._get_version_attachments(value.version)
            elif isinstance(value, DatasetVersionRepresentation):
                dataset_output = {
                    "useExistingDataset": True,
                    "useExistingVersion": True,
                    "datasetVersion": {"vecticeId": value.id, "name": value.name},
                }
                dataset_artifact = IterationStepArtifact(
                    datasetVersionId=value.id,
                    type=IterationStepArtifactType.DataSetVersion.name,
                )
                attachments_output = self._get_version_attachments(value)

            copy_artifacts = self.artifacts.copy()
            self.artifacts.append(dataset_artifact)
            attachments = (
                [
                    IterationStepArtifact(entityFileId=attach["fileId"], type="EntityFile")
                    for attach in attachments_output
                ]
                if attachments_output
                else []
            )
            self.artifacts += attachments
            self._keep_artifacts_and_update_step(StepType.StepDataset, value)
            step = StepDataset(self, dataset_artifact)
            _register_dataset_logging(self, dataset_output, value, attachments_output, _logger, copy_artifacts)
        else:
            raise TypeError(f"Expected Comment, Dataset or a Model, got {type(value)}")

        self._iteration._steps[self.name] = step

    def _check_string_sanity(self, value: str):
        if value == "":
            raise VecticeException("Cannot assign an empty comment. Please provide a valid comment")

    def _iadd_comment_or_image(self, value, artifact: IterationStepArtifact):
        self.artifacts.append(artifact)
        self._warn_step_change(value)
        return _comment_or_image_logging

    def _keep_artifacts_and_update_step(self, step_type: StepType, value) -> None:
        artifacts = list(
            filter(
                self._filter_non_empty_artifact,
                self._client.get_step_by_name(self.name, self._iteration.id).artifacts,
            )
        )

        def _get_artifact_id(artifact: IterationStepArtifact) -> int | str | None:
            if artifact.model_version_id:
                return artifact.model_version_id
            if artifact.dataset_version_id:
                return artifact.dataset_version_id
            if artifact.entity_file_id:
                return artifact.entity_file_id

            return None

        def _maintain_and_add_artifacts(session_artifacts, existing_artifacts):
            # Ensure we don't keep adding an instances artifacts to the step
            maintain_artifacts = existing_artifacts
            for ia, art in enumerate(session_artifacts):
                if len(maintain_artifacts) >= ia + 1:
                    existing_art = maintain_artifacts[ia]
                    if (art.id is not None and existing_art.id == art.id) or (
                        art.text is not None and existing_art.text == art.text
                    ):
                        continue
                    maintain_artifacts += [art]
                else:
                    maintain_artifacts += [art]
            return maintain_artifacts

        # These are artifacts that exist in the current instance
        filtered_artifacts = list(filter(self._filter_non_empty_artifact, self.artifacts))
        session_artifacts = [
            IterationStepArtifactInput(
                id=_get_artifact_id(artifact),
                type=artifact.type.value,
                text=str(artifact.text) if artifact.text is not None else None,
            )
            for artifact in filtered_artifacts
        ]
        # These are artifacts in Vectice
        existing_artifacts = [
            IterationStepArtifactInput(
                id=_get_artifact_id(artifact),
                type=artifact.type.value,
                text=str(artifact.text) if artifact.text is not None else None,
            )
            for artifact in artifacts
        ]
        # Ensure no crossover or duplicates
        maintain_artifacts = _maintain_and_add_artifacts(session_artifacts, existing_artifacts)
        self._client.update_iteration_step_artifact(self.id, step_type, artifacts=maintain_artifacts)
        self._warn_step_change(value)

    def _flush_artifacts(self, step_type: StepType):
        self._client.update_iteration_step_artifact(step_id=self.id, step_type=step_type, artifacts=[])
        self.artifacts = []

    def _step_factory_and_update(self, value) -> Step | StepString | StepNumber | StepDataset | StepModel:
        # TODO cyclic imports
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation
        from vectice.models.step_dataset import StepDataset
        from vectice.models.step_image import StepImage
        from vectice.models.step_model import StepModel
        from vectice.models.step_number import StepNumber
        from vectice.models.step_string import StepString

        self._ensure_correct_project_id_from_representation_objs(value)

        if isinstance(value, (int, float)):
            self._update_iteration_step(StepType.StepNumber, value)
            return StepNumber(step=self, number=value)

        img_value = is_image(value)
        if img_value:
            self._update_iteration_step(StepType.StepImage, img_value, True)
            return StepImage(self, img_value)

        if isinstance(value, str):
            self._check_string_sanity(value)
            self._update_iteration_step(StepType.StepString, value)
            return StepString(self, string=value)

        if (
            isinstance(value, Model)
            or isinstance(value, ModelRepresentation)
            or isinstance(value, ModelVersionRepresentation)
        ):
            data: dict[str, Any] = {}
            if isinstance(value, Model):
                code_version_id = self._get_code_version_id()
                data = self._client.register_model(
                    model=value,
                    project_id=self.project.id,
                    phase_id=self.phase.id,
                    iteration_id=self.iteration.id,
                    code_version_id=code_version_id,
                )
                attachments_output = self._set_model_attachments(value, data.model_version)
            elif isinstance(value, ModelRepresentation):
                data = {
                    "useExistingModel": True,
                    "modelVersion": {"vecticeId": value.version.id, "name": value.version.name},
                }
                attachments_output = self._get_version_attachments(value.version)
            elif isinstance(value, ModelVersionRepresentation):
                data = {
                    "useExistingModel": True,
                    "modelVersion": {"vecticeId": value.id, "name": value.name},
                }
                attachments_output = self._get_version_attachments(value)

            artifacts = self._get_version_artifacts(data, attachments_output)
            self._client.update_iteration_step_artifact(
                self.id,
                StepType.StepModel,
                artifacts,
            )
            _register_model_logging(self, data, value, self.name, attachments_output, _logger)
            self.artifacts = [
                IterationStepArtifact(modelVersionId=data["modelVersion"]["vecticeId"], type="ModelVersion")
            ]
            self._warn_step_change(value)
            return StepModel(self, self.artifacts[0], value)

        elif (
            isinstance(value, Dataset)
            or isinstance(value, DatasetRepresentation)
            or isinstance(value, DatasetVersionRepresentation)
        ):
            dataset: dict[str, Any] = {}
            if isinstance(value, Dataset):
                code_version_id = self._get_code_version_id()
                dataset = self._client.register_dataset_from_source(
                    dataset=value,
                    project_id=self.project.id,
                    phase_id=self.phase.id,
                    iteration_id=self.iteration.id,
                    code_version_id=code_version_id,
                )
                attachments_output = self._set_dataset_attachments(value, dataset.dataset_version)
            elif isinstance(value, DatasetRepresentation):
                dataset = {
                    "useExistingDataset": True,
                    "useExistingVersion": True,
                    "datasetVersion": {"vecticeId": value.version.id, "name": value.version.name},
                }
                attachments_output = self._get_version_attachments(value.version)
            elif isinstance(value, DatasetVersionRepresentation):
                dataset = {
                    "useExistingDataset": True,
                    "useExistingVersion": True,
                    "datasetVersion": {"vecticeId": value.id, "name": value.name},
                }
                attachments_output = self._get_version_attachments(value)

            artifacts = self._get_version_artifacts(dataset, attachments_output)
            self._client.update_iteration_step_artifact(
                self.id,
                StepType.StepDataset,
                artifacts,
            )
            _register_dataset_logging(self, dataset, value, attachments_output, _logger)
            self.artifacts = [
                IterationStepArtifact(datasetVersionId=dataset["datasetVersion"]["vecticeId"], type="DataSetVersion")
            ]
            self._warn_step_change(value)
            return StepDataset(self, dataset_version=self.artifacts[0])

        return self

    def _get_version_attachments(self, value: ModelVersionRepresentation | DatasetVersionRepresentation):
        return [
            vars(attach)
            for attach in self.project._client.list_version_representation_attachments(self.project.id, value).list
        ]

    def _ensure_correct_project_id_from_representation_objs(self, value):
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        if (
            isinstance(value, DatasetRepresentation)
            or isinstance(value, DatasetVersionRepresentation)
            or isinstance(value, ModelRepresentation)
            or isinstance(value, ModelVersionRepresentation)
        ) and value.project_id != self.project.id:
            raise VecticeException("Assigning an asset coming from another project is forbidden")

    def _get_version_artifacts(
        self, data, attachments_output: list[dict] | None = None
    ) -> list[IterationStepArtifactInput]:
        attachments = None
        if attachments_output:
            attachments = (
                [IterationStepArtifactInput(id=attach["fileId"], type="EntityFile") for attach in attachments_output]
                if attachments_output
                else None
            )
        if "modelVersion" in data:
            artifact = IterationStepArtifactInput(
                id=data["modelVersion"]["vecticeId"],
                type=IterationStepArtifactType.ModelVersion.name,
            )
        else:
            artifact = IterationStepArtifactInput(
                id=data["datasetVersion"]["vecticeId"],
                type=IterationStepArtifactType.DataSetVersion.name,
            )
        artifacts = [artifact]
        artifacts += attachments if attachments else []
        return artifacts

    def _create_image_artifact(
        self, value: str | IOBase | Image, flush_step_artifacts=False
    ) -> tuple[IterationStepArtifactInput, str]:
        if flush_step_artifacts is True:
            self._flush_artifacts(step_type=StepType.StepImage)

        image, filename = _get_image_variables(value)
        try:
            artifact, *_ = self._client.create_phase_attachments(
                [("file", (filename, image))], self.phase.id, self.project.id, self.id
            )
            if isinstance(image, BufferedReader):
                image.close()
            return (
                IterationStepArtifactInput(id=artifact["fileId"], type=IterationStepArtifactType.EntityFile.name),
                filename,
            )
        except Exception as error:
            raise ValueError("Check the provided image.") from error

    def _update_iteration_step(self, type: StepType, value, flush_step_artifacts=False):
        filename: str | None = None
        if type is StepType.StepImage:
            _, filename = self._create_image_artifact(value=value, flush_step_artifacts=flush_step_artifacts)
            self._warn_step_change(StepType.StepImage)
        elif isinstance(value, (str, int, float)):
            self._client.update_iteration_step_artifact(
                self.id,
                type,
                [IterationStepArtifactInput(type="Comment", text=str(value))],
            )
            self._warn_step_change(value)

        _comment_or_image_logging(self, _logger, filename)

    def _warn_step_change(self, value):
        # TODO: cyclic import
        from vectice.models.dataset import Dataset
        from vectice.models.model import Model
        from vectice.models.representation.dataset_representation import DatasetRepresentation
        from vectice.models.representation.dataset_version_representation import DatasetVersionRepresentation
        from vectice.models.representation.model_representation import ModelRepresentation
        from vectice.models.representation.model_version_representation import ModelVersionRepresentation

        if self._type is StepType.Step:
            return
        elif self._type is not StepType.StepModel and (
            isinstance(value, Model)
            or isinstance(value, ModelRepresentation)
            or isinstance(value, ModelVersionRepresentation)
        ):
            _logger.debug(f"Step type changed from {self._type.name} to StepModel")
        elif self._type is not StepType.StepDataset and (
            isinstance(value, Dataset)
            or isinstance(value, DatasetRepresentation)
            or isinstance(value, DatasetVersionRepresentation)
        ):
            _logger.debug(f"Step type changed from {self._type.name} to StepDataset")
        elif self._type is not StepType.StepImage and StepType.StepImage is value:
            _logger.debug(f"Step type changed from {self._type.name} to StepImage")
        elif self._type is not StepType.StepNumber and isinstance(value, (int, float)):
            _logger.debug(f"Step type changed from {self._type.name} to StepNumber")
        elif self._type is not StepType.StepString and isinstance(value, str):
            _logger.debug(f"Step type changed from {self._type.name} to StepString")

    @property
    def name(self) -> str:
        """The step's name.

        Returns:
            The step's name.
        """
        return self._name

    @property
    def id(self) -> int:
        """The step's id.

        Returns:
            The step's id.
        """
        return self._id

    @id.setter
    def id(self, step_id: int):
        """Set the step's id.

        Parameters:
            step_id: The id to set.
        """
        self._id = step_id

    @property
    def index(self) -> int:
        """The step's index.

        Returns:
            The step's index.
        """
        return self._index

    @property
    def slug(self) -> str:
        """The step's slug.

        Returns:
            The step's slug.
        """
        return self._slug

    @property
    def properties(self) -> dict:
        """The step's name, id, and index.

        Returns:
            A dictionary containing the `name`, `id` and `index` items.
        """
        return {"name": self.name, "id": self.id, "index": self.index}

    @property
    def artifacts(self) -> list[IterationStepArtifact]:
        return self._artifacts

    @artifacts.setter
    def artifacts(self, artifacts: list[IterationStepArtifact]):
        self._artifacts = artifacts

    @property
    def connection(self) -> Connection:
        """The connection to which this step belongs.

        Returns:
            The connection to which this step belongs.
        """
        return self._iteration.connection

    @property
    def workspace(self) -> Workspace:
        """The workspace to which this step belongs.

        Returns:
            The workspace to which this step belongs.
        """
        return self._iteration.workspace

    @property
    def project(self) -> Project:
        """The project to which this step belongs.

        Returns:
            The project to which this step belongs.
        """
        return self._iteration.project

    @property
    def phase(self) -> Phase:
        """The phase to which this step belongs.

        Returns:
            The phase to which this step belongs.
        """
        return self._iteration.phase

    @property
    def iteration(self) -> Iteration:
        """The iteration to which this step belongs.

        Returns:
            The iteration to which this step belongs.
        """
        return self._iteration

    def _get_code_version_id(self):
        # TODO: cyclic imports
        from vectice.models.git_version import _check_code_source, _inform_if_git_repo

        if vectice.code_capture:
            return _check_code_source(self._client, self.project.id, _logger)
        else:
            _inform_if_git_repo()
            return None

    @property
    def model(self) -> Model | None:
        """The step's model.

        Returns:
            The step's model.
        """
        return self._model

    def _set_model_attachments(self, model: Model, model_version: ModelVersionOutput):
        logging.getLogger("vectice.models.attachment_container").propagate = True
        attachments = None
        if model.attachments:
            container = AttachmentContainer(model_version, self._client)
            attachments = container.add_attachments(model.attachments)
        if model.predictor:
            model_content = self._serialize_model(model.predictor)
            model_type_name = type(model.predictor).__name__
            container = AttachmentContainer(model_version, self._client)
            container.add_serialized_model(model_type_name, model_content)
        return attachments

    def _set_dataset_attachments(self, dataset: Dataset, dataset_version: DatasetVersionOutput):
        logging.getLogger("vectice.models.attachment_container").propagate = True
        attachments = None
        if dataset.attachments:
            container = AttachmentContainer(dataset_version, self._client)
            attachments = container.add_attachments(dataset.attachments)
        return attachments

    def _filter_non_empty_artifact(self, artifact: IterationStepArtifact) -> str | int | float | None:
        return artifact.model_version_id or artifact.dataset_version_id or artifact.entity_file_id or artifact.text

    @staticmethod
    def _serialize_model(model: Any) -> bytes:
        return pickle.dumps(model)



from __future__ import annotations

from enum import Enum
from typing import TYPE_CHECKING

from vectice.api.json.metric import MetricInput

if TYPE_CHECKING:
    from vectice.api.json.model_version import ModelVersionOutput, ModelVersionStatus
    from vectice.models.property import Property


class ModelType(Enum):
    """Enumeration of the different model types."""

    ANOMALY_DETECTION = "ANOMALY_DETECTION"
    CLASSIFICATION = "CLASSIFICATION"
    CLUSTERING = "CLUSTERING"
    OTHER = "OTHER"
    RECOMMENDATION_MODELS = "RECOMMENDATION_MODELS"
    REGRESSION = "REGRESSION"
    TIME_SERIES = "TIME_SERIES"


class ModelRegisterOutput(dict):
    @property
    def model_version(self) -> ModelVersionOutput:
        from vectice.api.json.model_version import ModelVersionOutput

        return ModelVersionOutput(**self["modelVersion"])

    @property
    def use_existing_model(self) -> bool:
        return bool(self["useExistingModel"])

    # TODO: complete the jobrun property
    @property
    def job_run(self) -> str:
        return str(self["jobRun"])


class ModelRegisterInput(dict):
    @property
    def name(self) -> str:
        return str(self["name"])

    @property
    def model_type(self) -> ModelType:
        return ModelType(**self["modelType"])

    @property
    def properties(self) -> Property:
        from vectice.models.property import Property

        return Property(**self["properties"])

    @property
    def metrics(self) -> list[MetricInput]:
        return [MetricInput(**metric) for metric in self["metrics"]]

    @property
    def status(self) -> ModelVersionStatus:
        from vectice.api.json.model_version import ModelVersionStatus

        return ModelVersionStatus(self["status"])

    @property
    def framework(self) -> str:
        return str(self["framework"])

    @property
    def type(self) -> str:
        return str(self["type"])

    @property
    def algorithm_name(self) -> str:
        return str(self["algorithmName"])

    @property
    def uri(self) -> str:
        return str(self["uri"])

    @property
    def derived_from(self) -> list[int]:
        return [(self["inputs"])]



INDEX_ORDERED = {"field": "index", "direction": "ASC"}
INDEX_ORDERED_DESC = {"field": "index", "direction": "DESC"}
PAGINATE_OUTPUT = """
            items {{
                {0}
            }}
            page {{
                index
                size
            }}
            total
"""


def get_page_input(index=1, size=10) -> dict:
    return {"index": index, "size": size}



from vectice.api.json.dataset_version_representation import DatasetVersionRepresentationOutput


class DatasetVersionRepresentation:
    def __init__(self, output: DatasetVersionRepresentationOutput):
        self.id = output.id
        self.project_id = output.project_id
        self.name = output.name
        self.description = output.description



from __future__ import annotations

from enum import Enum


class VersionStrategy(Enum):
    """Enumeration of the supported version strategies."""

    AUTOMATIC = "AUTOMATIC"
    MANUAL = "MANUAL"


class ArtifactVersion(dict):
    def __init__(
        self,
        version_number: int | None = None,
        version_name: str | None = None,
        version_id: str | None = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        if version_number is not None:
            self["versionNumber"] = version_number
        if version_name is not None:
            self["versionName"] = version_name
        if version_id is not None:
            self["id"] = version_id

    @property
    def version_number(self) -> int | None:
        return self.get("versionNumber", None)

    @property
    def version_id(self) -> int | None:
        return self.get("id", None)

    @property
    def version_name(self) -> int | None:
        return self.get("versionName", None)
